# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import logging
import os

from flask import current_app
from flask_script import Manager
from jinja2 import Environment, Markup, PackageLoader

from .base import log_config, logger

#: sub-manager for config commands
manager = Manager(
    description='Show config / create default config',
    help='Show config / create default config')


@manager.command
def show(only_path=False):
    """Show the current config.
    """
    logger.setLevel(logging.INFO)
    infos = [
        '\n',
        'Instance path: "{}"'.format(current_app.instance_path),
        'CONFIG_ENVVAR: "{}"'.format(current_app.CONFIG_ENVVAR),
    ]

    logger.info('\n  '.join(infos))

    if not only_path:
        log_config(current_app.config)


class DefaultConfig(object):
    SQLALCHEMY_DATABASE_URI = ''

    PRODUCTION = False
    LOGGING_CONFIG_FILE = 'logging.yml'

    # SQLALCHEMY_DATABASE_URI = 'sqlite:///:memory:'
    SQLALCHEMY_ECHO = False
    REDIS_URI = 'redis://localhost:6379/1'

    DEBUG = True
    ASSETS_DEBUG = True
    DEBUG_TB_ENABLED = True
    TEMPLATE_DEBUG = False
    DEBUG_TB_INTERCEPT_REDIRECTS = False
    DEBUG_TB_PROFILER_ENABLED = False

    BROKER_URL = 'redis://localhost:6379/1'
    CELERY_RESULT_BACKEND = 'redis://localhost:6379/1'

    CELERYD_PREFETCH_MULTIPLIER = 1
    CELERY_ALWAYS_EAGER = True  # run tasks locally, no async
    CELERY_EAGER_PROPAGATES_EXCEPTIONS = True

    WHOOSH_BASE = 'whoosh'

    SITE_NAME = u''
    MAIL_SENDER = u''
    UNOCONV_LOCATION = u'/usr/bin/unoconv'

    def __init__(self, logging_file=None):
        self.SESSION_COOKIE_NAME = '{}-session'.format(current_app.name)
        self.SECRET_KEY = os.urandom(24)

        db_dir = os.path.join(current_app.instance_path, 'data')
        if not os.path.exists(db_dir):
            os.mkdir(db_dir)
        self.SQLALCHEMY_DATABASE_URI = \
          "sqlite:///{}/data/db.sqlite".format(current_app.instance_path)

        if logging_file:
            self.LOGGING_CONFIG_FILE = logging_file


class ReprProxy(object):
    """Proxy an object and apply repr() + Mark safe when accesing an attribute.

    Used in jinja templates.
    """

    def __init__(self, obj):
        self.__obj = obj

    def __getattr__(self, name):
        return Markup(repr(self.__obj.__getattribute__(name)))


def write_config(config_file, config):
    jinja_env = Environment(loader=PackageLoader(__name__, 'templates'))
    template = jinja_env.get_template('config.py.jinja2')

    with open(config_file, 'w') as f:
        f.write(template.render(cfg=ReprProxy(config)))
    logger.info('Generated "%s"', config_file)


def maybe_write_logging(logging_file):
    if not os.path.exists(logging_file):
        jinja_env = Environment(loader=PackageLoader(__name__, 'templates'))
        template = jinja_env.get_template('logging.yml.jinja2')

        with open(logging_file, 'w') as f:
            f.write(template.render())
        logger.info('Generated "%s"', logging_file)

    else:
        logger.info(
            'Logging config file "%s" already exists, skipping creation.',
            logging_file)


@manager.command
def init(filename='config.py', logging_config='logging.yml'):
    """Create default config files in instance folder.

    * [FILENAME] (default: "config.py")
    * [LOGGING_CONFIG] (default: logging.yml)

    Defaults are tailored for development.
    """
    app = current_app
    config_file = os.path.join(app.instance_path, filename)
    logging_file = os.path.join(app.instance_path, logging_config)

    if os.path.exists(config_file):
        logger.info('Config file  "%s" already exists! Abort.', config_file)
        return 1

    config = DefaultConfig(logging_file=logging_config)
    write_config(config_file, config)
    maybe_write_logging(logging_file)



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from abilian.core.singleton import UniqueName


class NS1(UniqueName):
    pass


class NS2(UniqueName):
    pass


def test_singleton():
    val = NS1('val')
    other_val = NS1('val')
    assert val is other_val
    assert id(val) == id(other_val)


def test_equality():
    val = NS1('val')
    assert val == 'val'
    assert val == u'val'


def test_namespaces():
    ns1_val = NS1('val')
    ns2_val = NS2('val')
    assert ns1_val is not ns2_val
    # equality works because of string compat
    assert ns1_val == ns2_val



# coding=utf-8
"""
Subject classes (i.e. people, groups, etc.).

See ICOM-ics-v1.0 "Subject Branch".

TODO: I'm not a big fan of the "subject" name. Could be replaced by something
else, like "people" or "principal" ?
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import random
import string
from abc import ABCMeta, abstractmethod, abstractproperty
from datetime import datetime, timedelta

import bcrypt
import sqlalchemy as sa
from flask_login import UserMixin, current_app
from six import python_2_unicode_compatible, text_type
from sqlalchemy.event import listens_for
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import backref, deferred, relationship
from sqlalchemy.schema import Column, ForeignKey, Table, UniqueConstraint
from sqlalchemy.types import Boolean, DateTime, Integer, LargeBinary, \
    UnicodeText

from abilian.core import sqlalchemy as sa_types

from .base import SEARCHABLE, SYSTEM, IdMixin, Indexable, TimestampedMixin, db

__all__ = ['User', 'Group', 'Principal']

# Tables for many-to-many relationships
following = Table(
    'following',
    db.Model.metadata,
    Column('follower_id', Integer, ForeignKey('user.id')),
    Column('followee_id', Integer, ForeignKey('user.id')),
    UniqueConstraint('follower_id', 'followee_id'),)
membership = Table(
    'membership',
    db.Model.metadata,
    Column(
        'user_id',
        Integer,
        ForeignKey(
            'user.id', onupdate='CASCADE', ondelete='CASCADE')),
    Column(
        'group_id',
        Integer,
        ForeignKey(
            'group.id', onupdate='CASCADE', ondelete='CASCADE')),
    UniqueConstraint('user_id', 'group_id'),)

# Should not be needed (?)
administratorship = Table(
    'administratorship',
    db.Model.metadata,
    Column(
        'user_id',
        Integer,
        ForeignKey(
            'user.id', onupdate='CASCADE', ondelete='CASCADE')),
    Column(
        'group_id',
        Integer,
        ForeignKey(
            'group.id', onupdate='CASCADE', ondelete='CASCADE')),
    UniqueConstraint('user_id', 'group_id'),)

_RANDOM_PASSWORD_CHARS = (
    string.ascii_letters + string.digits + string.punctuation)


def gen_random_password(length=15):
    rg = random.SystemRandom()
    return u''.join(rg.choice(_RANDOM_PASSWORD_CHARS) for i in range(length))


class PasswordStrategy(object):

    __metaclass__ = ABCMeta

    @abstractproperty
    def name(self):
        """Strategy name.
        """

    @abstractmethod
    def authenticate(self, user, password):
        """Predicate to tell wether password match user's or not.
        """

    @abstractmethod
    def process(self, user, password):
        """Return a string to be stored as user password
        """


class ClearPasswordStrategy(PasswordStrategy):
    """Don't encrypt at all.

    This strategy should not ever be used elsewhere than in tests. It's useful
    in tests since a hash like bcrypt is designed to be slow.
    """

    @property
    def name(self):
        return "clear"

    def authenticate(self, user, password):
        return user.password == password

    def process(self, user, password):
        if not isinstance(password, text_type):
            password = password.decode('utf-8')
        return password


class BcryptPasswordStrategy(PasswordStrategy):
    """Hash passwords using bcrypt.
    """

    @property
    def name(self):
        return 'bcrypt'

    def authenticate(self, user, password):
        current_passwd = user.password
        # crypt work only on str, not unicode
        if isinstance(current_passwd, text_type):
            current_passwd = current_passwd.encode('utf-8')
        if isinstance(password, text_type):
            password = password.encode('utf-8')

        return bcrypt.hashpw(password, current_passwd) == current_passwd

    def process(self, user, password):
        if isinstance(password, text_type):
            password = password.encode('utf-8')
        return bcrypt.hashpw(password, bcrypt.gensalt()).decode('utf-8')


class UserQuery(db.Model.query_class):

    def get_by_email(self, email):
        return self.filter_by(email=email).one()


class Principal(IdMixin, TimestampedMixin, Indexable):
    """A principal is either a User or a Group."""

    __indexation_args__ = {}
    __indexation_args__.update(Indexable.__indexation_args__)
    index_to = __indexation_args__.setdefault('index_to', ())
    __indexation_args__['index_to'] += (('name',
                                         ('name', 'name_prefix', 'text')),)
    del index_to

    def has_role(self, role):
        return current_app.services['security'].has_role(self, role)


@python_2_unicode_compatible
class User(Principal, UserMixin, db.Model):
    __tablename__ = 'user'
    __editable__ = ['first_name', 'last_name', 'email', 'password']
    __exportable__ = __editable__ + ['created_at', 'updated_at', 'id']

    __password_strategy__ = BcryptPasswordStrategy()

    entity_type = u'{}.{}'.format(__module__, 'User')

    query_class = UserQuery

    # Basic information
    first_name = Column(UnicodeText, info=SEARCHABLE)
    last_name = Column(UnicodeText, info=SEARCHABLE)
    # Should we add gender, salutation ?

    # System information
    email = Column(UnicodeText, nullable=False)
    can_login = Column(Boolean, nullable=False, default=True)
    password = Column(
        UnicodeText, default="*", info={'audit_hide_content': True})

    photo = deferred(Column(LargeBinary))

    last_active = Column(DateTime, info=SYSTEM)
    locale = Column(sa_types.Locale, nullable=True, default=None)
    timezone = Column(sa_types.Timezone, nullable=True, default=None)

    __table_args__ = (UniqueConstraint('email'),)

    followers = relationship(
        "User",
        secondary=following,
        primaryjoin=('User.id == following.c.follower_id'),
        secondaryjoin=('User.id == following.c.followee_id'),
        backref='followees')

    def __init__(self, password=None, **kwargs):
        Principal.__init__(self)
        UserMixin.__init__(self)
        db.Model.__init__(self, **kwargs)

        if self.can_login and password is not None:
            self.set_password(password)

    def authenticate(self, password):
        if self.password and self.password != "*":
            return self.__password_strategy__.authenticate(self, password)
        else:
            return False

    def set_password(self, password):
        """Encrypts and sets password."""
        self.password = self.__password_strategy__.process(self, password)

    def follow(self, followee):
        if followee == self:
            raise Exception("User can't follow self")
        self.followees.append(followee)

    def unfollow(self, followee):
        if followee == self:
            raise Exception("User can't follow self")
        i = self.followees.index(followee)
        del self.followees[i]

    def join(self, group):
        self.groups.add(group)

    def leave(self, group):
        if group in self.groups:
            self.groups.remove(group)

    #
    # Boolean properties
    #
    def is_following(self, other):
        return other in self.followees

    def is_member_of(self, group):
        return self in group.members

    def is_admin_of(self, group):
        return self in group.admins

    @property
    def is_online(self):
        if self.last_active is None:
            return False
        return datetime.utcnow() - self.last_active <= timedelta(minutes=1)

    #
    # Other properties
    #
    @property
    def name(self):
        name = u'{first_name} {last_name}'.format(
            first_name=self.first_name or u'', last_name=self.last_name or u'')
        return name.strip() or u'Unknown'

    @property
    def short_name(self):
        first_name = self.first_name or u''
        last_name = self.last_name[0:1] + "." if self.last_name else u''
        name = u'{} {}'.format(first_name, last_name)
        return name.strip() or u'Unknown'

    def __str__(self):
        return self.name

    def __repr__(self):
        cls = self.__class__
        return '<{mod}.{cls} id={id!r} email={email!r} at 0x{addr:x}>'.format(
            mod=cls.__module__,
            cls=cls.__name__,
            id=self.id,
            email=self.email,
            addr=id(self))


@listens_for(User, "mapper_configured", propagate=True)
def _add_user_indexes(mapper, class_):
    # this is a functional index (indexes on a function result), we cannot define
    # it in __table_args__.
    #
    # see: https://groups.google.com/d/msg/sqlalchemy/CgSJUlelhGs/_Nj3f201hs4J
    idx = sa.schema.Index(
        'user_unique_lowercase_email',
        sa.sql.func.lower(class_.email),
        unique=True)
    idx.info['engines'] = ('postgresql',)


class Group(Principal, db.Model):
    __indexable__ = False
    __tablename__ = 'group'
    __editable__ = ['name', 'description']
    __exportable__ = __editable__ + ['created_at', 'updated_at', 'id']

    entity_type = u'{}.{}'.format(__module__, 'Group')

    name = Column(UnicodeText, nullable=False, info=SEARCHABLE)
    description = Column(UnicodeText, info=SEARCHABLE)

    members = relationship(
        "User",
        collection_class=set,
        secondary=membership,
        backref=backref(
            'groups', lazy='select', collection_class=set))
    admins = relationship(
        "User", collection_class=set, secondary=administratorship)

    photo = deferred(Column(LargeBinary))

    public = Column(Boolean, default=False, nullable=False)

    @hybrid_property
    def members_count(self):
        return len(self.members)

    @members_count.expression
    def members_count(cls):
        return sa.sql \
            .select([sa.sql.func.count(membership.c.user_id)]) \
            .where(membership.c.group_id == cls.id) \
            .group_by(membership.c.group_id) \
            .label('members_count')



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import shutil
import weakref
from pathlib import Path
from uuid import UUID, uuid1

import six
import sqlalchemy as sa
from flask import _app_ctx_stack, current_app
from flask.globals import _lookup_app_object
from six import string_types, text_type
from sqlalchemy.orm.session import Session

from abilian.services import Service, ServiceState

_NULL_MARK = object()


def _assert_uuid(uuid):
    if not isinstance(uuid, UUID):
        raise ValueError('Not an uuid.UUID instance', uuid)


class RepositoryServiceState(ServiceState):
    #: :class:`Path` path to application repository
    path = None  # type: Path


class RepositoryService(Service):
    """
    Service for storage of binary objects referenced in database
    """
    name = 'repository'
    AppStateClass = RepositoryServiceState

    def init_app(self, app):
        Service.init_app(self, app)

        path = app.DATA_DIR / 'files'
        if not path.exists():
            path.mkdir(0o775)

        with app.app_context():
            self.app_state.path = path.resolve()

    # data management: paths and accessors
    def rel_path(self, uuid):
        """Contruct relative path from repository top directory to the file
        named after this uuid.

        :param:uuid: :class:`UUID` instance
        """
        _assert_uuid(uuid)
        filename = str(uuid)
        return Path(filename[0:2], filename[2:4], filename)

    def abs_path(self, uuid):
        """Return absolute :class:`Path` object for given uuid.

        :param:uuid: :class:`UUID` instance
        """
        top = self.app_state.path
        rel_path = self.rel_path(uuid)
        dest = top / rel_path
        assert top in dest.parents
        return dest

    def get(self, uuid, default=None):
        """
        Return absolute :class:`Path` object for given uuid, if this uuid exists in
        repository, or `default` if it doesn't.

        :param:uuid: :class:`UUID` instance
        """
        path = self.abs_path(uuid)
        if not path.exists():
            return default
        return path

    def set(self, uuid, content, encoding='utf-8'):
        """Store binary content with uuid as key

        :param:uuid: :class:`UUID` instance
        :param:content: string, bytes, or any object with a `read()` method
        :param:encoding: encoding to use when content is Unicode
        """
        dest = self.abs_path(uuid)
        if not dest.parent.exists():
            dest.parent.mkdir(0o775, parents=True)

        if hasattr(content, 'read'):
            content = content.read()

        mode = 'tw'
        if not isinstance(content, text_type):
            mode = 'bw'
            encoding = None

        with dest.open(mode, encoding=encoding) as f:
            f.write(content)

    def delete(self, uuid):
        """Delete file with given uuid.

        :param:uuid: :class:`UUID` instance
        :raises:KeyError if file does not exists
        """
        dest = self.abs_path(uuid)
        if not dest.exists():
            raise KeyError('No file can be found for this uuid', uuid)

        dest.unlink()

    def __getitem__(self, uuid):
        v = self.get(uuid, default=_NULL_MARK)
        if v is _NULL_MARK:
            raise KeyError('No file can be found for this uuid', uuid)
        return v

    def __setitem__(self, uuid, content):
        self.set(uuid, content)

    def __delitem__(self, uuid):
        self.delete(uuid)


repository = RepositoryService()

_REPOSITORY_TRANSACTION = 'abilian_repository_transactions'


class SessionRepositoryState(ServiceState):
    path = None  # type: Path

    @property
    def transactions(self):
        try:
            return _lookup_app_object(_REPOSITORY_TRANSACTION)
        except AttributeError:
            reg = dict()
            setattr(_app_ctx_stack.top, _REPOSITORY_TRANSACTION, reg)
            return reg

    @transactions.setter
    def transactions(self, value):
        top = _app_ctx_stack.top
        if top is None:
            raise RuntimeError('working outside of application context')

        setattr(top, _REPOSITORY_TRANSACTION, value)

    # transaction <-> db session accessors
    def get_transaction(self, session):
        if isinstance(session, sa.orm.scoped_session):
            session = session()

        s_id = id(session)
        default = (weakref.ref(session), None)
        s_ref, transaction = self.transactions.get(s_id, default)
        s = s_ref()

        if s is None or s is not session:
            # old session with same id, maybe not yet garbage collected
            transaction = None
        return transaction

    def set_transaction(self, session, transaction):
        """
        :param:session: :class:`sqlalchemy.orm.session.Session` instance
        :param:transaction: :class:`RepositoryTransaction` instance
        """
        if isinstance(session, sa.orm.scoped_session):
            session = session()

        s_id = id(session)
        self.transactions[s_id] = (weakref.ref(session), transaction)

    def create_transaction(self, session, transaction):
        if not self.running:
            return

        parent = self.get_transaction(session)
        root_path = self.path
        transaction = RepositoryTransaction(root_path, parent)
        self.set_transaction(session, transaction)

    def end_transaction(self, session, transaction):
        if not self.running:
            return

        tr = self.get_transaction(session)
        if tr is not None:
            if not tr.cleared:
                # root and nested transactions emit "commit", but subtransactions don't
                tr.commit(session)
            self.set_transaction(session, tr._parent)

    def begin(self, session):
        if not self.running:
            return

        tr = self.get_transaction(session)
        if tr is None:
            # FIXME: return or create a new one?
            return

        return tr.begin(session)

    def commit(self, session):
        if not self.running:
            return

        tr = self.get_transaction(session)
        if tr is None:
            return

        tr.commit(session)

    def flush(self, session, flush_context):
        # when sqlalchemy is flushing it is done in a sub-transaction, not the root
        # one. So when calling our 'commit' from here we are not in our root
        # transaction, so changes will not be written to repository.
        self.commit(session)

    def rollback(self, session):
        if not self.running:
            return

        tr = self.get_transaction(session)
        if tr is None:
            return

        tr.rollback(session)


class SessionRepositoryService(Service):
    """
    A repository service that is session aware, i.e content is actually
    written or delete at commit time.

    All content is stored using the main :class:`RepositoryService`.
    """
    name = 'session_repository'
    AppStateClass = SessionRepositoryState

    def __init__(self, *args, **kwargs):
        self.__listening = False
        Service.__init__(self, *args, **kwargs)

    def init_app(self, app):
        Service.init_app(self, app)

        path = Path(app.instance_path, 'tmp', 'files_transactions')
        if not path.exists():
            path.mkdir(0o775, parents=True)

        with app.app_context():
            self.app_state.path = path.resolve()

        if not self.__listening:
            self.__listening = True
            listen = sa.event.listen
            listen(Session, "after_transaction_create", self.create_transaction)
            listen(Session, "after_transaction_end", self.end_transaction)
            listen(Session, "after_begin", self.begin)
            listen(Session, "after_commit", self.commit)
            listen(Session, "after_flush", self.flush)
            listen(Session, "after_rollback", self.rollback)
            # appcontext_tearing_down.connect(self.clear_transaction, app)

    def _session_for(self, model_or_session):
        """Return session instance for object parameter.

        If parameter is a session instance, it is return as is.
        If parameter is a registered model instance, its session will be used.

        If parameter is a detached model instance, or None, application scoped
        session will be used (app.db.session())

        If parameter is a scoped_session instance, a new session will be
        instanciated.
        """
        session = model_or_session
        if not isinstance(session, (Session, sa.orm.scoped_session)):
            if session is not None:
                session = sa.orm.object_session(model_or_session)

            if session is None:
                session = current_app.db.session

        if isinstance(session, sa.orm.scoped_session):
            session = session()

        return session

    # repository interface
    def get(self, session, uuid, default=None):
        session = self._session_for(session)
        tr = self.app_state.get_transaction(session)
        try:
            val = tr.get(uuid)
        except KeyError:
            return default

        if val is _NULL_MARK:
            val = repository.get(uuid, default)

        return val

    def set(self, session, uuid, content, encoding='utf-8'):
        session = self._session_for(session)
        tr = self.app_state.get_transaction(session)
        tr.set(uuid, content, encoding)

    def delete(self, session, uuid):
        session = self._session_for(session)
        tr = self.app_state.get_transaction(session)
        if self.get(session, uuid) is not None:
            tr.delete(uuid)

    # session event handlers
    @Service.if_running
    def create_transaction(self, session, transaction):
        return self.app_state.create_transaction(session, transaction)

    @Service.if_running
    def end_transaction(self, session, transaction):
        return self.app_state.end_transaction(session, transaction)

    @Service.if_running
    def begin(self, session, transaction, connection):
        return self.app_state.begin(session)

    @Service.if_running
    def commit(self, session):
        return self.app_state.commit(session)

    @Service.if_running
    def flush(self, session, flush_context):
        return self.app_state.flush(session, flush_context)

    @Service.if_running
    def rollback(self, session):
        return self.app_state.rollback(session)


session_repository = SessionRepositoryService()


class RepositoryTransaction(object):

    def __init__(self, root_path, parent=None):
        self.path = root_path / str(uuid1())
        # if parent is not None and parent.cleared:
        #   parent = None

        self._parent = parent
        self._deleted = set()
        self._set = set()
        self.__cleared = False

    @property
    def cleared(self):
        return self.__cleared

    def __del__(self):
        if not self.cleared:
            self._clear()

    def _clear(self):
        if self.__cleared:
            return

        # make sure transaction is not usable anymore
        if self.path.exists():
            shutil.rmtree(str(self.path))

        del self.path
        del self._deleted
        del self._set
        self.__cleared = True

    def begin(self, session=None):
        if not self.path.exists():
            self.path.mkdir(0o700)

    def rollback(self, session=None):
        self._clear()

    def commit(self, session=None):
        """
        Merge modified objects into parent transaction.

        Once commited a transaction object is not usable anymore

        :param:session: current sqlalchemy Session
        """
        if self.__cleared:
            return

        if self._parent:
            # nested transaction
            self._commit_parent()
        else:
            self._commit_repository()
        self._clear()

    def _commit_repository(self):
        assert self._parent is None

        for uuid in self._deleted:
            try:
                repository.delete(uuid)
            except KeyError:
                pass

        for uuid in self._set:
            content = self.path / str(uuid)
            repository.set(uuid, content.open('rb'))

    def _commit_parent(self):
        p = self._parent
        p._deleted |= self._deleted
        p._deleted -= self._set

        p._set |= self._set
        p._set -= self._deleted

        if self._set:
            p.begin()  # ensure p.path exists

        for uuid in self._set:
            content_path = self.path / str(uuid)
            # content_path.replace is not available with python < 3.3.
            content_path.rename(p.path / str(uuid))

    def _add_to(self, uuid, dest, other):
        """
        Add `item` to `dest` set, ensuring `item` is not present in `other` set.
        """
        _assert_uuid(uuid)
        try:
            other.remove(uuid)
        except KeyError:
            pass
        dest.add(uuid)

    def delete(self, uuid):
        self._add_to(uuid, self._deleted, self._set)

    def set(self, uuid, content, encoding='utf-8'):
        self.begin()
        self._add_to(uuid, self._set, self._deleted)

        if hasattr(content, 'read'):
            content = content.read()

        if isinstance(content, bytes):
            mode = 'bw'
            encoding = None
        else:
            mode = 'tw'

        dest = self.path / str(uuid)
        with dest.open(mode, encoding=encoding) as f:
            f.write(content)

    def get(self, uuid):
        if uuid in self._deleted:
            raise KeyError

        if uuid in self._set:
            path = self.path / str(uuid)
            assert path.exists()
            return path

        if self._parent:
            return self._parent.get(uuid)

        return _NULL_MARK



"""Tag service.

Manages tags applied to taggable entities.

Cf. ATOM specs  (+ for instance this discussion: http://edward.oconnor.cx/2007/02/representing-tags-in-atom)
Cf. ICOM 3.6.12 (Tag) et 3.6.13 (TagApplication.

Warning: the Activity Steams spec has a notion of tagging, but it's slightly
different than what we are used to in content management:
"Indicates that the actor has associated the object with the target.
For example, if the actor specifies that a particular user appears in a photo.
The object is the user and the target is the photo."

"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from sqlalchemy import Column, DateTime, ForeignKey, Integer, Unicode

from abilian.core.entities import IdMixin
from abilian.core.extensions import db


class Tag(db.Model, IdMixin):
    """Persistent class representing a tag."""

    #: Unique, autogenerated id.
    id = Column(Integer, primary_key=True)

    #: The text associated to the tag
    term = Column(Unicode, nullable=False, unique=True)

    # Not sure we want the extra complexity, even though it's specified by Atom.
    #label = Column()

    # Also we might want: created_at, etc. Maybe a Tag should be an Entity?


class TagApplication(db.Model, IdMixin):
    """Persistent class representing a tag application on a taggable object.
    """

    #: Unique, autogenerated id.
    id = Column(Integer, primary_key=True)

    #: The actor that did the tagging.
    actor_id = Column(Integer, ForeignKey("user.id"))

    #: The entity (object) on which the tag has been applied.
    entity_id = Column(Integer)  # must point to a valid entity

    #: The time when the tag was applied.
    created_at = Column(DateTime, nullable=False)



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, division

import pkg_resources

from flask import url_for, current_app
from flask_assets import Bundle

from abilian.services.security import Anonymous

# register custom filters for webassets
from . import filters  # noqa


def init_app(app):
    assets = app.extensions['webassets']
    assets.append_path(RESOURCES_DIR, '/static/abilian')
    app.add_static_url(
        'abilian', RESOURCES_DIR, endpoint='abilian_static', roles=Anonymous)

    app.before_first_request(requirejs_config)


def requirejs_config():
    assets = current_app.extensions['webassets']
    config = assets.requirejs_config

    # setup ckeditor
    ckeditor_lib = 'ckeditor/ckeditor'
    config['shim']['ckeditor'] = {'exports': 'CKEDITOR'}
    config['paths']['ckeditor'] = url_for(
        'abilian_static', filename=ckeditor_lib)

    d3_lib = 'nvd3/d3.min'
    config['shim']['d3'] = {'exports': 'd3'}
    config['paths']['d3'] = url_for('abilian_static', filename=d3_lib)

    nvd3_lib = 'nvd3/nv.d3'
    config['shim']['nvd3'] = {'exports': 'nv', 'deps': ['d3']}
    config['paths']['nvd3'] = url_for('abilian_static', filename=nvd3_lib)


RESOURCES_DIR = pkg_resources.resource_filename('abilian.web', 'resources')

JQUERY = Bundle('jquery/js/jquery-1.11.3.js')

BOOTBOX_JS = Bundle('bootbox/bootbox.js')

BOOTSTRAP_LESS = Bundle('bootstrap/less/bootstrap.less')
BOOTSTRAP_JS = Bundle('bootstrap/js/bootstrap.js')

BOOTSTRAP_DATEPICKER_LESS = 'bootstrap-datepicker/less/datepicker.less'
BOOTSTRAP_DATEPICKER_JS = Bundle(
    'bootstrap-datepicker/js/bootstrap-datepicker.js')

BOOTSTRAP_SWITCH_LESS = Bundle(
    'bootstrap-switch/less/bootstrap3/bootstrap-switch.less')
BOOTSTRAP_SWITCH_JS = Bundle('bootstrap-switch/bootstrap-switch.js')

BOOTSTRAP_TIMEPICKER_LESS = Bundle('bootstrap-timepicker/less/timepicker.less')
BOOTSTRAP_TIMEPICKER_JS = Bundle(
    'bootstrap-timepicker/js/bootstrap-timepicker.js')

DATATABLE_LESS = Bundle('datatables/css/jquery.dataTables.css',
                        'datatables/css/jquery.dataTables_themeroller.css')
DATATABLE_JS = Bundle('datatables/js/jquery.dataTables.js')

FILEAPI_JS = Bundle('fileapi/FileAPI.js')

FONTAWESOME_LESS = Bundle('font-awesome/less/font-awesome.less')

REQUIRE_JS = Bundle('requirejs/require.js', 'requirejs/domReady.js')

SELECT2_LESS = Bundle(
    'select2/select2.css',
    'select2/select2-bootstrap.css',)
SELECT2_JS = Bundle('select2/select2.js')

TYPEAHEAD_LESS = Bundle('typeahead/typeahead.js-bootstrap.less')
TYPEAHEAD_JS = Bundle('typeahead/typeahead.js', 'typeahead/hogan-2.0.0.js')

ABILIAN_LESS = Bundle('less/abilian.less', 'less/print.less')

ABILIAN_JS_NS = Bundle('js/abilian-namespace.js')
ABILIAN_JS = Bundle(
    'js/abilian.js',
    'js/datatables-setup.js',
    'js/datatables-advanced-search.js',
    'js/widgets/base.js',
    'js/widgets/select2.js',
    'js/widgets/richtext.js',
    'js/widgets/delete.js',
    'js/widgets/file.js',
    'js/widgets/image.js',
    'js/widgets/tags.js',
    'js/widgets/dynamic-row.js',)

LESS = Bundle(
    BOOTSTRAP_LESS,
    FONTAWESOME_LESS,
    SELECT2_LESS,
    TYPEAHEAD_LESS,
    BOOTSTRAP_DATEPICKER_LESS,
    BOOTSTRAP_SWITCH_LESS,
    BOOTSTRAP_TIMEPICKER_LESS,
    DATATABLE_LESS,
    ABILIAN_LESS,)

TOP_JS = Bundle(REQUIRE_JS, JQUERY, ABILIAN_JS_NS)

JS = Bundle(
    BOOTSTRAP_JS,
    TYPEAHEAD_JS,
    BOOTBOX_JS,
    SELECT2_JS,
    BOOTSTRAP_DATEPICKER_JS,
    BOOTSTRAP_SWITCH_JS,
    BOOTSTRAP_TIMEPICKER_JS,
    DATATABLE_JS,
    FILEAPI_JS,
    ABILIAN_JS,)

JS_I18N = (
    'select2/select2_locale_{lang}.js',
    'bootstrap-datepicker/js/locales/bootstrap-datepicker.{lang}.js',)



"""
Conversion service.

Hardcoded to manage only conversion to PDF, to text and to image series.

Includes result caching (on filesystem).

Assumes poppler-utils and LibreOffice are installed.

TODO: rename Converter into ConversionService ?
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import glob
import hashlib
import logging
import mimetypes
import os
import re
import shutil
import subprocess
import threading
import traceback
from abc import ABCMeta, abstractmethod
from base64 import decodestring, encodestring
from contextlib import contextmanager
from io import BytesIO
from pathlib import Path
from tempfile import mkstemp

import six
from magic import Magic
from PIL import Image
from PIL.ExifTags import TAGS
from six import raise_from, string_types, text_type

from abilian.services.image import FIT, resize

try:
    from xmlrpclib import ServerProxy
except ImportError:
    ServerProxy = None

logger = logging.getLogger(__name__)

# For some reason, twill includes a broken version of subprocess.
assert 'twill' not in subprocess.__file__

# Hack for Mac OS + homebrew
os.environ['PATH'] += ":/usr/local/bin"

TMP_DIR = "tmp"
CACHE_DIR = "cache"


def get_tmp_dir():
    return converter.TMP_DIR


class ConversionError(Exception):
    pass


class HandlerNotFound(ConversionError):
    pass


class Cache(object):

    CACHE_DIR = None

    def _path(self, key):
        """File path for `key`:
        """
        return self.CACHE_DIR / "{}.blob".format(key)

    def __contains__(self, key):
        return self._path(key).exists()

    def get(self, key):
        if key in self:
            value = self._path(key).open('rb').read()
            if key.startswith("txt:"):
                value = text_type(value, encoding="utf8")
            return value
        else:
            return None

    __getitem__ = get

    def set(self, key, value):
        # if not os.path.exists(self.CACHE_DIR):
        #   os.mkdir(CACHE_DIR)
        fd = self._path(key).open("wb")
        if key.startswith("txt:"):
            fd.write(value.encode("utf8"))
        else:
            fd.write(value)
        fd.close()

    __setitem__ = set

    def clear(self):
        pass


class Converter(object):

    def __init__(self):
        self.handlers = []
        self.cache = Cache()

    def init_app(self, app):
        self.init_work_dirs(
            cache_dir=Path(app.instance_path, CACHE_DIR),
            tmp_dir=Path(app.instance_path, TMP_DIR),)

        app.extensions['conversion'] = self

        for handler in self.handlers:
            handler.init_app(app)

    def init_work_dirs(self, cache_dir, tmp_dir):
        self.TMP_DIR = Path(tmp_dir)
        self.CACHE_DIR = Path(cache_dir)
        self.cache.CACHE_DIR = self.CACHE_DIR

        if not self.TMP_DIR.exists():
            self.TMP_DIR.mkdir()
        if not self.CACHE_DIR.exists():
            self.CACHE_DIR.mkdir()

    def clear(self):
        self.cache.clear()
        for d in (self.TMP_DIR, self.CACHE_DIR):
            shutil.rmtree(bytes(d))
            d.mkdir()

    def register_handler(self, handler):
        self.handlers.append(handler)

    # TODO: refactor, pass a "File" or "Document" or "Blob" object
    def to_pdf(self, digest, blob, mime_type):
        cache_key = "pdf:" + digest
        pdf = self.cache.get(cache_key)
        if pdf:
            return pdf

        for handler in self.handlers:
            if handler.accept(mime_type, "application/pdf"):
                pdf = handler.convert(blob)
                self.cache[cache_key] = pdf
                return pdf
        raise HandlerNotFound("No handler found to convert from %s to PDF" %
                              mime_type)

    def to_text(self, digest, blob, mime_type):
        """Convert a file to plain text.

        Useful for full-text indexing. Returns a Unicode string.
        """
        # Special case, for now (XXX).
        if mime_type.startswith("image/"):
            return u""

        cache_key = "txt:" + digest

        text = self.cache.get(cache_key)
        if text:
            return text

        # Direct conversion possible
        for handler in self.handlers:
            if handler.accept(mime_type, "text/plain"):
                text = handler.convert(blob)
                self.cache[cache_key] = text
                return text

        # Use PDF as a pivot format
        pdf = self.to_pdf(digest, blob, mime_type)
        for handler in self.handlers:
            if handler.accept("application/pdf", "text/plain"):
                text = handler.convert(pdf)
                self.cache[cache_key] = text
                return text

        raise HandlerNotFound("No handler found to convert from %s to text" %
                              mime_type)

    def has_image(self, digest, mime_type, index, size=500):
        """Tell if there is a preview image.
        """
        cache_key = "img:%s:%s:%s" % (index, size, digest)
        return mime_type.startswith("image/") or cache_key in self.cache

    def get_image(self, digest, blob, mime_type, index, size=500):
        """Return an image for the given content, only if it already exists in
        the image cache.
        """
        # Special case, for now (XXX).
        if mime_type.startswith("image/"):
            return ""

        cache_key = "img:%s:%s:%s" % (index, size, digest)
        return self.cache.get(cache_key)

    def to_image(self, digest, blob, mime_type, index, size=500):
        """Convert a file to a list of images. Returns image at the given index.
        """
        # Special case, for now (XXX).
        if mime_type.startswith("image/"):
            return ""

        cache_key = "img:%s:%s:%s" % (index, size, digest)
        converted = self.cache.get(cache_key)
        if converted:
            return converted

        # Direct conversion possible
        for handler in self.handlers:
            if handler.accept(mime_type, "image/jpeg"):
                converted_images = handler.convert(blob, size=size)
                for i in range(0, len(converted_images)):
                    converted = converted_images[i]
                    self.cache["img:%s:%s:%s" % (i, size, digest)] = converted
                return converted_images[index]

        # Use PDF as a pivot format
        pdf = self.to_pdf(digest, blob, mime_type)
        for handler in self.handlers:
            if handler.accept("application/pdf", "image/jpeg"):
                converted_images = handler.convert(pdf, size=size)
                for i in range(0, len(converted_images)):
                    converted = converted_images[i]
                    self.cache["img:%s:%s:%s" % (i, size, digest)] = converted
                return converted_images[index]

        raise HandlerNotFound("No handler found to convert from %s to image" %
                              mime_type)

    def get_metadata(self, digest, content, mime_type):
        """Get a dictionary representing the metadata embedded in the given
        content.
        """

        # XXX: ad-hoc for now, refactor later
        if mime_type.startswith("image/"):
            img = Image.open(BytesIO(content))
            ret = {}
            if not hasattr(img, '_getexif'):
                return {}
            info = img._getexif()
            if not info:
                return {}
            for tag, value in info.items():
                decoded = TAGS.get(tag, tag)
                ret["EXIF:" + str(decoded)] = value
            return ret
        else:
            if mime_type != "application/pdf":
                content = self.to_pdf(digest, content, mime_type)

            with make_temp_file(content) as in_fn:
                try:
                    output = subprocess.check_output(['pdfinfo', in_fn])
                except OSError:
                    logger.error(
                        "Conversion failed, probably pdfinfo is not installed")
                    raise

            ret = {}
            for line in output.split(b"\n"):
                if b":" in line:
                    key, value = line.strip().split(b":", 1)
                    ret["PDF:" + key] = text_type(
                        value.strip(), errors="replace")

            return ret

    @staticmethod
    def digest(blob):
        assert isinstance(blob, string_types)
        if isinstance(blob, str):
            digest = hashlib.md5(blob).hexdigest()
        else:
            digest = hashlib.md5(blob.encode("utf8")).hexdigest()
        return digest


class Handler(object):
    __metaclass__ = ABCMeta

    accepts_mime_types = []
    produces_mime_types = []

    def __init__(self, *args, **kwargs):
        self.log = logger.getChild(self.__class__.__name__)

    def init_app(self, app):
        pass

    @property
    def TMP_DIR(self):
        return get_tmp_dir()

    @property
    def mime_sniffer(self):
        return Magic(mime=True)

    @property
    def encoding_sniffer(self):
        return Magic(mime_encoding=True)

    def accept(self, source_mime_type, target_mime_type):
        """Generic matcher based on patterns."""

        match_source = False
        match_target = False

        for pat in self.accepts_mime_types:
            if re.match("^%s$" % pat, source_mime_type):
                match_source = True
                break

        for pat in self.produces_mime_types:
            if re.match("^%s$" % pat, target_mime_type):
                match_target = True
                break

        return match_source and match_target

    @abstractmethod
    def convert(self, key, **kw):
        pass


class PdfToTextHandler(Handler):
    accepts_mime_types = ['application/pdf', 'application/x-pdf']
    produces_mime_types = ['text/plain']

    def convert(self, blob, **kw):
        with make_temp_file(blob) as in_fn, make_temp_file() as out_fn:
            try:
                subprocess.check_call(['pdftotext', in_fn, out_fn])
            except Exception as e:
                raise raise_from(ConversionError('pdftotext failed'), e)

            converted = open(out_fn).read()
            encoding = self.encoding_sniffer.from_file(out_fn)

        if encoding in ("binary", None):
            encoding = "ascii"
        try:
            converted_unicode = text_type(converted, encoding, errors="ignore")
        except:
            traceback.print_exc()
            converted_unicode = text_type(converted, errors="ignore")

        return converted_unicode


class AbiwordTextHandler(Handler):
    accepts_mime_types = ['application/msword']
    produces_mime_types = ['text/plain']

    def convert(self, blob, **kw):
        tmp_dir = self.TMP_DIR
        cur_dir = os.getcwd()
        with make_temp_file(blob, suffix=".doc") as in_fn,\
             make_temp_file(suffix='.txt') as out_fn:
            try:
                os.chdir(str(tmp_dir))
                subprocess.check_call([
                    'abiword', '--to', os.path.basename(out_fn),
                    os.path.basename(in_fn)
                ])
            except Exception as e:
                raise_from(ConversionError('abiword failed'), e)
            finally:
                os.chdir(cur_dir)

            converted = open(out_fn).read()
            encoding = self.encoding_sniffer.from_file(out_fn)

        if encoding in ("binary", None):
            encoding = "ascii"
        try:
            converted_unicode = text_type(converted, encoding, errors="ignore")
        except:
            traceback.print_exc()
            converted_unicode = text_type(converted, errors="ignore")

        return converted_unicode


class AbiwordPDFHandler(Handler):
    accepts_mime_types = [
        'application/msword',
        'application/vnd.oasis.opendocument.text',
        'text/rtf',
    ]
    produces_mime_types = ['application/pdf']

    def convert(self, blob, **kw):
        cur_dir = os.getcwd()
        with make_temp_file(blob, suffix=".doc") as in_fn,\
             make_temp_file(suffix='.pdf') as out_fn:
            try:
                os.chdir(bytes(self.TMP_DIR))
                subprocess.check_call([
                    'abiword', '--to', os.path.basename(out_fn),
                    os.path.basename(in_fn)
                ])
            except Exception as e:
                raise_from(ConversionError('abiword failed'), e)
            finally:
                os.chdir(cur_dir)

            converted = open(out_fn).read()
            return converted


class ImageMagickHandler(Handler):
    accepts_mime_types = ['image/.*']
    produces_mime_types = ['application/pdf']

    def convert(self, blob, **kw):
        with make_temp_file(blob) as in_fn, make_temp_file() as out_fn:
            try:
                subprocess.check_call(['convert', in_fn, "pdf:" + out_fn])
                converted = open(out_fn, 'rb').read()
                return converted
            except Exception as e:
                raise_from(ConversionError('convert failed'), e)


class PdfToPpmHandler(Handler):
    accepts_mime_types = ['application/pdf', 'application/x-pdf']
    produces_mime_types = ['image/jpeg']

    def convert(self, blob, size=500):
        """Size is the maximum horizontal size."""
        l = []
        with make_temp_file(blob) as in_fn, make_temp_file() as out_fn:
            try:
                subprocess.check_call(['pdftoppm', '-jpeg', in_fn, out_fn])
                l = glob.glob("%s-*.jpg" % out_fn)
                l.sort()

                converted_images = []
                for fn in l:
                    converted = resize(
                        open(fn, 'rb').read(), size, size, mode=FIT)
                    converted_images.append(converted)

                return converted_images
            except Exception as e:
                raise_from(ConversionError('pdftoppm failed'), e)
            finally:
                for fn in l:
                    try:
                        os.remove(fn)
                    except OSError:
                        pass


class UnoconvPdfHandler(Handler):
    """Handles conversion from office documents (MS-Office, OOo) to PDF.

    Uses unoconv.
    """

    # TODO: add more if needed.
    accepts_mime_types = [
        'application/vnd.oasis.*', 'application/msword',
        'application/mspowerpoint', 'application/vnd.ms-powerpoint',
        'application/vnd.ms-excel', 'application/ms-excel',
        'application/vnd.openxmlformats-officedocument.*', 'text/rtf'
    ]
    produces_mime_types = ['application/pdf']
    run_timeout = 60
    _process = None
    unoconv = 'unoconv'

    def init_app(self, app):
        unoconv = app.config.get('UNOCONV_LOCATION')
        found = False
        execute_ok = False

        if unoconv:
            found = os.path.isfile(unoconv)
            if found:
                # make absolute path: avoid errors when running with different CWD
                unoconv = os.path.abspath(unoconv)
                execute_ok = os.access(unoconv, os.X_OK)
                if not execute_ok:
                    self.log.warning('Not allowed to execute "{}", fallback to '
                                     '"unoconv"'.format(unoconv))
            else:
                self.log.warning('Cannot find "{}", fallback to "unoconv"'
                                 ''.format(unoconv))
        if (not unoconv or not found or not execute_ok):
            unoconv = 'unoconv'

        self.unoconv = unoconv

    @property
    def unoconv_version(self):
        # Hack for my Mac, FIXME later
        if os.path.exists(
                "/Applications/LibreOffice.app/Contents/program/python"):
            cmd = [
                '/Applications/LibreOffice.app/Contents/program/python',
                '/usr/local/bin/unoconv', '--version'
            ]
        else:
            cmd = [self.unoconv, '--version']

        process = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        out, err = process.communicate()
        return out

    def convert(self, blob, **kw):
        """Convert using unoconv converter.
        """
        timeout = self.run_timeout
        with make_temp_file(blob) as in_fn, \
                make_temp_file(prefix='tmp-unoconv-', suffix=".pdf") as out_fn:

            # Hack for my Mac, FIXME later
            if os.path.exists(
                    "/Applications/LibreOffice.app/Contents/program/python"):
                cmd = [
                    '/Applications/LibreOffice.app/Contents/program/python',
                    '/usr/local/bin/unoconv', '-f', 'pdf', '-o', out_fn, in_fn
                ]
            else:
                cmd = [self.unoconv, '-f', 'pdf', '-o', out_fn, in_fn]

            def run_uno():
                try:
                    self._process = subprocess.Popen(
                        cmd, close_fds=True, cwd=bytes(self.TMP_DIR))
                    self._process.communicate()
                except Exception as e:
                    logger.error('run_uno error: %s', bytes(e), exc_info=True)
                    raise_from(ConversionError('unoconv failed'), e)

            run_thread = threading.Thread(target=run_uno)
            run_thread.start()
            run_thread.join(timeout)

            try:
                if run_thread.is_alive():
                    # timeout reached
                    self._process.terminate()
                    if self._process.poll() is not None:
                        try:
                            self._process.kill()
                        except OSError:
                            logger.warning("Failed to kill process {}".format(
                                self._process))

                    self._process = None
                    raise ConversionError("Conversion timeout ({})".format(
                        timeout))

                converted = open(out_fn).read()
                return converted
            finally:
                self._process = None


class CloudoooPdfHandler(Handler):
    """Handles conversion from OOo to PDF.

    Highly inefficient since file are serialized in base64 over HTTP.

    Deactivated because it's so hard to set up.

    FIXME: needs cleanup, or removal.
    """

    accepts_mime_types = [r'application/.*']
    produces_mime_types = ['application/pdf']

    # Hardcoded for now
    SERVER_URL = "http://localhost:8011"

    pivot_format_map = {
        "doc": "odt",
        "docx": "odt",
        "xls": "ods",
        "xlsx": "ods",
        "ppt": "odp",
        "pptx": "odp",
    }

    def convert(self, key):
        in_fn = "data/%s.blob" % key
        in_mime_type = open("data/%s.mime" % key).read()
        file_extension = mimetypes.guess_extension(in_mime_type).strip(".")

        data = encodestring(open(in_fn).read())
        proxy = ServerProxy(self.SERVER_URL, allow_none=True)

        if in_mime_type.startswith("application/vnd.oasis.opendocument"):
            data = proxy.convertFile(data, file_extension, 'pdf')
        else:
            pivot_format = self.pivot_format_map[file_extension]
            data = proxy.convertFile(data, file_extension, pivot_format)
            data = proxy.convertFile(data, pivot_format, 'pdf')

        converted = decodestring(data)
        new_key = hashlib.md5(converted).hexdigest()
        with open("data/%s.blob" % new_key, "wb") as fd:
            fd.write(converted)
        return new_key


class WvwareTextHandler(Handler):
    accepts_mime_types = ['application/msword']
    produces_mime_types = ['text/plain']

    def convert(self, blob, **kw):

        with make_temp_file(blob) as in_fn, make_temp_file() as out_fn:
            try:
                subprocess.check_call(['wvText', in_fn, out_fn])
            except Exception as e:
                raise_from(ConversionError('wxText failed'), e)

            converted = open(out_fn).read()

            encoding = self.encoding_sniffer.from_file(out_fn)
            if encoding in ("binary", None):
                encoding = "ascii"
            try:
                converted_unicode = text_type(
                    converted, encoding, errors="ignore")
            except:
                traceback.print_exc()
                converted_unicode = text_type(converted, errors="ignore")

            return converted_unicode


# Utils
@contextmanager
def make_temp_file(blob=None, prefix='tmp', suffix="", tmp_dir=None):
    if tmp_dir is None:
        tmp_dir = get_tmp_dir()

    fd, filename = mkstemp(dir=str(tmp_dir), prefix=prefix, suffix=suffix)
    if blob is not None:
        fd = os.fdopen(fd, 'wb')
        fd.write(blob)
        fd.close()
    else:
        os.close(fd)

    yield filename
    try:
        os.remove(filename)
    except OSError:
        pass


# Singleton, yuck!
converter = Converter()
converter.register_handler(PdfToTextHandler())
converter.register_handler(PdfToPpmHandler())
converter.register_handler(ImageMagickHandler())

_unoconv_handler = UnoconvPdfHandler()
converter.register_handler(_unoconv_handler)

#converter.register_handler(AbiwordPDFHandler())
#converter.register_handler(AbiwordTextHandler())

# Needs to be rewriten
#converter.register_handler(CloudoooPdfHandler())



"""
Activity Service.

See: http://activitystrea.ms/specs/json/1.0/
See: http://activitystrea.ms/specs/atom/1.0/#activity
See: http://stackoverflow.com/questions/1443960/how-to-implement-the-activity-stream-in-a-social-network

TODO: Look wether other attributes from the spec need to be implemented.
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from datetime import datetime

from flask import logging
from six import text_type
from sqlalchemy.orm import relationship
from sqlalchemy.schema import Column, ForeignKey
from sqlalchemy.types import DateTime, Integer, String, Text

from abilian.core.entities import Entity, db
from abilian.core.models.subjects import User

__all__ = ['ActivityEntry']

logger = logging.getLogger(__name__)


def _default_from(column):
    """
    Helper for default and onupdates parameters in a Column definitions.

    Returns a `context-sensitive default function <http://docs.sqlalchemy.org/en/rel_0_8/core/defaults.html#context-sensitive-default-functions>`_
    to set value from another column.
    """

    def _default_value(context):
        return context.current_parameters[column]

    return _default_value


class ActivityEntry(db.Model):
    """
    Main table for all activities.
    """

    id = Column(Integer, primary_key=True)
    happened_at = Column(DateTime, default=datetime.utcnow)

    verb = Column(Text)

    actor_id = Column(Integer, ForeignKey(User.id))
    actor = relationship(User, foreign_keys=actor_id)

    object_type = Column(String(1000))
    object_id = Column(Integer, default=_default_from('_fk_object_id'))
    _fk_object_id = Column(Integer, ForeignKey(Entity.id, ondelete="SET NULL"))
    object = relationship(Entity, foreign_keys=_fk_object_id)

    target_type = Column(String(1000))
    target_id = Column(Integer, default=_default_from('_fk_target_id'))
    _fk_target_id = Column(Integer, ForeignKey(Entity.id, ondelete="SET NULL"))
    target = relationship(Entity, foreign_keys=_fk_target_id)

    def __repr__(self):
        return ('<{}.ActivityEntry id={} actor={} verb={} object={} '
                'target={}>'.format(self.__class__.__module__, self.id,
                                    repr(text_type(self.actor)),
                                    repr(self.verb),
                                    repr(text_type(self.object)),
                                    repr(text_type(self.target))))



from __future__ import absolute_import, division, print_function, \
    unicode_literals


class PreferencePanel(object):
    """Base class for preference panels.

    Currently, this class does nothing. I may be useful in the future either
    as just a marker interface (for automatic plugin discovery / registration),
    or to add some common functionnalities. Otherwise, it will be removed.
    """

    def is_accessible(self):
        return True



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from datetime import datetime

import sqlalchemy as sa
from flask_login import current_user
from six import text_type
from werkzeug.exceptions import BadRequest

from abilian.core.entities import Entity
from abilian.core.models.comment import Comment, is_commentable
from abilian.core.util import utc_dt
from abilian.i18n import _, _l
from abilian.web import nav, url_for
from abilian.web.action import ButtonAction, actions
from abilian.web.blueprints import Blueprint
from abilian.web.views.object import CANCEL_BUTTON, ObjectCreate, \
    ObjectDelete, ObjectEdit

from .forms import CommentForm

bp = Blueprint('comments', __name__, url_prefix='/comments')


def _default_comment_view(obj, obj_type, obj_id, **kwargs):
    entity = obj.entity
    return url_for(entity, _anchor='comment-{}'.format(obj.id))


@bp.record_once
def register_default_view(state):
    state.app.default_view.register(Comment, _default_comment_view)


COMMENT_BUTTON = ButtonAction(
    'form', 'edit', btn_class='primary', title=_l(u'Post'))


class BaseCommentView(object):
    Model = Comment
    Form = CommentForm

    #: commented entity
    entity = None

    def init_object(self, args, kwargs):
        args, kwargs = super(BaseCommentView, self).init_object(args, kwargs)
        entity_id = kwargs.pop('entity_id', None)
        if entity_id is not None:
            self.entity = Entity.query.get(entity_id)

        if self.entity is None:
            raise BadRequest('No entity to comment')

        if not is_commentable(self.entity):
            raise BadRequest('This entity is not commentable')

        actions.context['object'] = self.entity
        return args, kwargs

    def view_url(self):
        kw = {}
        if self.obj and self.obj.id:
            kw['_anchor'] = 'comment-{}'.format(self.obj.id)
        return url_for(self.entity, **kw)

    def index_url(self):
        return self.view_url()

    @property
    def activity_target(self):
        return self.entity


class CommentEditView(BaseCommentView, ObjectEdit):

    _message_success = _l(u'Comment edited')

    def breadcrumb(self):
        label = _(u'Edit comment on "{title}"').format(title=self.entity.name)
        return nav.BreadcrumbItem(label=label)

    def get_form_buttons(self, *args, **kwargs):
        return [COMMENT_BUTTON, CANCEL_BUTTON]

    def after_populate_obj(self):
        obj_meta = self.obj.meta.setdefault('abilian.core.models.comment', {})
        history = obj_meta.setdefault('history', [])
        history.append(
            dict(
                user_id=current_user.id,
                user=text_type(current_user),
                date=utc_dt(datetime.utcnow()).isoformat(),))
        self.obj.meta.changed()


edit_view = CommentEditView.as_view('edit')
bp.route('/<int:entity_id>/<int:object_id>/edit')(edit_view)


class CommentCreateView(BaseCommentView, ObjectCreate):
    """
  """
    _message_success = _l(u"Comment added")

    def __init__(self, *args, **kwargs):
        super(CommentCreateView, self).__init__(*args, **kwargs)

    def init_object(self, args, kwargs):
        args, kwargs = super(CommentCreateView, self).init_object(args, kwargs)
        self.obj.entity = self.entity
        session = sa.orm.object_session(self.entity)

        if session:
            sa.orm.session.make_transient(self.obj)

        return args, kwargs

    def breadcrumb(self):
        label = _(u'New comment on "{title}"').format(title=self.entity.name)
        return nav.BreadcrumbItem(label=label)

    def get_form_buttons(self, *args, **kwargs):
        return [COMMENT_BUTTON, CANCEL_BUTTON]


create_view = CommentCreateView.as_view('create')
bp.route('/<int:entity_id>/create')(create_view)


class CommentDeleteView(BaseCommentView, ObjectDelete):

    _message_success = _l(u"Comment deleted")


delete_view = CommentDeleteView.as_view('delete')
bp.route('/<int:entity_id>/<int:object_id>/delete')(delete_view)



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, unicode_literals

from flask import current_app
from flask_debugtoolbar.panels import DebugPanel

from abilian.core.entities import Entity
from abilian.core.models.subjects import Group
from abilian.i18n import _
from abilian.web.action import actions

from .models import Anonymous


class SecurityInfoDebugPanel(DebugPanel):
    """
    A panel to display current roles and permissions for "current" object.
    """
    name = 'SecurityInfo'

    @property
    def current_obj(self):
        return actions.context.get('object')

    @property
    def has_content(self):
        obj = self.current_obj
        return (obj is not None and isinstance(obj, Entity) and
                obj.id is not None)

    def nav_title(self):
        return _('Security Info')

    def nav_subtitle(self):
        """Subtitle showing until title in toolbar"""
        obj = self.current_obj
        if not obj:
            return _(u'No current object')

        try:
            return u'{}(id={})'.format(obj.__class__.__name__, obj.id)
        except:
            return u''

    def title(self):
        return self.nav_title()

    def url(self):
        return ''

    def content(self):
        obj = self.current_obj
        svc = current_app.services['security']
        context = self.context.copy()

        context['permissions'] = svc.get_permissions_assignments(obj=obj)
        context['roles'] = roles = dict()

        for principal, r in svc.get_role_assignements(obj=obj):
            if r not in roles:
                roles[r] = dict(anonymous=False, users=set(), groups=set())

            info = roles[r]
            if principal is Anonymous:
                info['anonymous'] = True
            elif isinstance(principal, Group):
                info['groups'].add(principal)
            else:
                info['users'].add(principal)

        for r in roles:
            info = roles[r]
            info['groups'] = [
                u'{g} (id={g.id})'.format(g=g)
                for g in sorted(
                    info['groups'], key=lambda g: g.name)
            ]
            users = sorted(
                info['users'],
                key=lambda u: (u.last_name.lower(), u.first_name.lower()))
            info['users'] = [
                u'{u} (id={u.id}, email="{u.email}")'.format(u=u) for u in users
            ]

        jinja_env = current_app.jinja_env
        jinja_env.filters.update(self.jinja_env.filters)
        template = jinja_env.get_or_select_template(
            'debug_panels/security_info_panel.html')

        return template.render(context)



"""
Test the index service.
"""
from __future__ import absolute_import, print_function, unicode_literals

from sqlalchemy import Column, Text, UnicodeText
from sqlalchemy.orm import column_property

from abilian.core.entities import SEARCHABLE, Entity
from abilian.services import index_service

from .base import IntegrationTestCase


def gen_name(ctx):
    params = ctx.current_parameters
    return u'{} {}'.format(
        params.get('first_name') or u'', params.get('last_name') or u'').strip()


class DummyContact1(Entity):
    name = column_property(
        Column(
            'name',
            UnicodeText(),
            info=SEARCHABLE,
            default=gen_name,
            onupdate=gen_name),
        Entity.name)

    salutation = Column(UnicodeText, default="")
    first_name = Column(UnicodeText, default="", info=SEARCHABLE)
    last_name = Column(UnicodeText, default="", info=SEARCHABLE)
    email = Column(Text, default="")


class IndexingTestCase(IntegrationTestCase):
    SERVICES = ('security', 'indexing')

    def tearDown(self):
        if index_service.running:
            index_service.stop()
        IntegrationTestCase.tearDown(self)

    def test_contacts_are_indexed(self):
        self.login_system()
        contact = DummyContact1(
            first_name="John", last_name="Test User", email="test@example.com")
        self.session.add(contact)
        self.session.commit()

        search_result = index_service.search(u'john')
        assert len(search_result) == 1
        found = search_result[0]
        assert contact.id == found['id']
        assert contact.name == found['name']

        search_result = index_service.search(u"john")
        assert len(search_result) == 1
        assert contact.id == int(search_result[0]['id'])



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import abc

import six
import sqlalchemy as sa
from sqlalchemy import Column, ForeignKey, Integer, UnicodeText
from sqlalchemy.orm import backref, relationship

from abilian.core.entities import Entity
from abilian.services.security import CREATE, DELETE, WRITE, Anonymous, Owner

#: name of backref on target :class:`Entity` object
ATTRIBUTE = '__comments__'


@six.add_metaclass(abc.ABCMeta)
class Commentable(object):
    pass


def register(cls):
    """Register an :class:`Entity` as a commentable class.

    Can be used as a class decorator:

    .. code-block:: python

      @comment.register
      class MyContent(Entity):
          ...
    """
    if not issubclass(cls, Entity):
        raise ValueError(
            'Class must be a subclass of abilian.core.entities.Entity')

    Commentable.register(cls)
    return cls


def is_commentable(obj):
    """
    :param obj: a class or instance
    """
    if isinstance(obj, type):
        return issubclass(obj, Commentable)

    if not isinstance(obj, Commentable):
        return False

    if obj.id is None:
        return False

    return True


def for_entity(obj, check_commentable=False):
    """Return comments on an entity.
    """
    if check_commentable and not is_commentable(obj):
        return []

    return getattr(obj, ATTRIBUTE)


class Comment(Entity):
    """A Comment related to an :class:`Entity`.
    """
    __default_permissions__ = {
        WRITE: {Owner},
        DELETE: {Owner},
        CREATE: {Anonymous},
    }

    @sa.ext.declarative.declared_attr
    def __mapper_args__(cls):
        # we cannot use super(Comment, cls): declared_attr happens during class
        # construction. super(cls, cls) could work; as long as `cls` is not a
        # subclass of `Comment`: it would enter into an infinite loop.
        #
        # Entity.__mapper_args__ calls the descriptor with 'Entity', not `cls`.
        args = Entity.__dict__['__mapper_args__'].fget(cls)
        args['order_by'] = cls.created_at
        return args

    entity_id = Column(Integer, ForeignKey(Entity.id), nullable=False)

    #: Commented entity
    entity = relationship(
        Entity,
        lazy='immediate',
        foreign_keys=[entity_id],
        backref=backref(
            ATTRIBUTE,
            lazy='select',
            order_by='Comment.created_at',
            cascade="all, delete-orphan",))

    #: comment's main content
    body = Column(
        UnicodeText(), sa.CheckConstraint("trim(body) != ''"), nullable=False)

    @property
    def history(self):
        return self.meta.get('abilian.core.models.comment', {}).get('history',
                                                                    [])

    def __repr__(self):
        class_ = self.__class__
        mod_ = class_.__module__
        classname = class_.__name__
        return '<{}.{} instance at 0x{:x} entity id={!r}'\
            .format(mod_, classname, id(self), self.entity_id)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import unittest

import pytest

from abilian.core.entities import Entity
from abilian.core.extensions import db
from abilian.core.models.subjects import Group, User
from abilian.testing import BaseTestCase

from . import READ, WRITE, Admin, Anonymous, Authenticated, Creator, \
    InheritSecurity, Owner, Permission, PermissionAssignment, Reader, Role, \
    RoleAssignment, SecurityAudit, Writer, security

TEST_EMAIL = u"joe@example.com"
TEST_PASSWORD = "tototiti"

skip = pytest.mark.skip


def init_user():
    user = User(
        first_name="Joe",
        last_name="User",
        email=TEST_EMAIL,
        password=TEST_PASSWORD)
    db.session.add(user)
    db.session.flush()


class RoleTestCase(unittest.TestCase):

    def test_singleton(self):
        admin = Role('admin')
        other_admin = Role('admin')
        self.assertIs(admin, other_admin)
        self.assertEqual(id(admin), id(other_admin))

    def test_equality(self):
        admin = Role('admin')
        self.assertEqual(admin, 'admin')
        self.assertEqual(admin, u'admin')

    def test_ordering(self):
        roles = [Authenticated, Admin, Anonymous]
        roles.sort()
        assert roles == [Admin, Anonymous, Authenticated]

    def test_enumerate_assignables(self):
        assert Role.assignable_roles() == [Admin]


class IntegrationTestCase(BaseTestCase):

    def setUp(self):
        BaseTestCase.setUp(self)
        # init_user()
        security.init_app(self.app)
        security.start()

    def tearDown(self):
        security.stop()
        security.clear()
        BaseTestCase.tearDown(self)


class DummyModel(Entity):
    pass


class FolderishModel(Entity, InheritSecurity):
    pass


class SecurityTestCase(IntegrationTestCase):

    def test_anonymous_user(self):
        # anonymous user is not an SQLAlchemy instance and must be handled
        # specifically to avoid tracebacks
        anon = self.app.login_manager.anonymous_user()
        assert not security.has_role(anon, 'reader')
        assert security.has_role(anon, Anonymous)
        assert security.get_roles(anon) == [Anonymous]
        assert not security.has_permission(anon, 'read')

    def test_has_role_authenticated(self):
        anon = self.app.login_manager.anonymous_user()
        user = User(email="john@example.com", password="x")
        self.session.add(user)
        self.session.flush()
        assert not security.has_role(anon, Authenticated)
        assert security.has_role(user, Authenticated)

    def test_root_user(self):
        """Root user always has any role, any permission.
        """
        root = User.query.get(0)
        assert security.has_role(root, Admin)
        assert security.has_role(root, Anonymous)
        assert security.has_permission(root, 'manage')

        obj = DummyModel()
        self.session.add(obj)
        self.session.flush()
        assert security.has_role(root, Admin, obj)
        assert security.has_permission(root, 'manage', obj)

    def test_grant_basic_roles(self):
        user = User(email="john@example.com", password="x")
        self.session.add(user)
        self.session.flush()

        # everybody always has role Anonymous
        assert security.has_role(user, Anonymous)

        security.grant_role(user, Admin)
        assert security.has_role(user, Admin)
        assert security.get_roles(user) == [Admin]
        assert security.get_roles(user) == ['admin']
        assert security.get_principals(Admin) == [user]

        # clear roles cache for better coverage: has_permission uses
        # _fill_role_cache_batch(), get_roles uses _fill_role_cache()
        delattr(user, '__roles_cache__')
        assert security.has_permission(user, "read")
        assert security.has_permission(user, "write")
        assert security.has_permission(user, "manage")

        security.ungrant_role(user, "admin")
        assert not security.has_role(user, "admin")
        assert security.get_roles(user) == []
        assert security.get_principals(Admin) == []

        assert not security.has_permission(user, "read")
        assert not security.has_permission(user, "write")
        assert not security.has_permission(user, "manage")

    def test_grant_basic_roles_on_groups(self):
        user = User(email="john@example.com", password="x")
        group = Group(name="Test Group")
        user.groups.add(group)
        self.session.add(user)
        self.session.flush()

        security.grant_role(group, "admin")
        assert security.has_role(group, "admin")
        assert security.get_roles(group) == ['admin']
        assert security.get_principals(Admin) == [group]

        assert security.has_role(user, Admin)

        assert security.has_permission(user, "read")
        assert security.has_permission(user, "write")
        assert security.has_permission(user, "manage")

        security.ungrant_role(group, "admin")
        assert not security.has_role(group, "admin")
        assert security.get_roles(group) == []
        assert security.get_principals(Admin) == []

        assert not security.has_role(user, "admin")
        assert not security.has_permission(user, "read")
        assert not security.has_permission(user, "write")
        assert not security.has_permission(user, "manage")

    def test_grant_roles_on_objects(self):
        user = User(email="john@example.com", password="x")
        user2 = User(email="papa@example.com", password="p")
        group = Group(name="Test Group")
        user.groups.add(group)
        obj = DummyModel()
        self.session.add_all([user, user2, obj])
        self.session.flush()

        security.grant_role(user, 'global_role')
        security.grant_role(user, "reader", obj)
        assert security.has_role(user, "reader", obj)
        assert security.get_roles(user, obj) == ['reader']
        assert security.get_principals(Reader) == []
        assert security.get_principals(Reader, object=obj) == [user]

        assert security.has_permission(user, "read", obj)
        assert not security.has_permission(user, "write", obj)
        assert not security.has_permission(user, "manage", obj)

        # test get_roles "global": object roles should not appear
        assert security.get_roles(user) == ['global_role']

        # global role is valid on all object
        assert security.has_role(user, 'global_role', obj)

        security.ungrant_role(user, "reader", obj)
        assert not security.has_role(user, "reader", obj)
        assert security.get_roles(user, obj) == []
        assert security.has_role(user, 'global_role', obj)

        assert not security.has_permission(user, "read", obj)
        assert not security.has_permission(user, "write", obj)
        assert not security.has_permission(user, "manage", obj)

        # owner / creator roles
        assert security.get_principals(Owner, object=obj) == []
        assert security.get_principals(Creator, object=obj) == []
        old_owner = obj.owner
        old_creator = obj.creator
        obj.owner = user
        assert security.get_roles(user, obj) == [Owner]
        assert security.get_principals(Owner, object=obj) == [user]
        assert security.get_principals(Creator, object=obj) == []
        # if user2 has Admin role e gets the rights no matter Creator/Ownership
        security.grant_role(user2, Admin)
        assert security.has_role(user2, (Owner, Creator), obj)
        assert security.has_role(user, (Owner, Creator), obj)

        obj.owner = old_owner
        obj.creator = user
        assert security.get_roles(user, obj) == [Creator]
        assert security.get_principals(Owner, object=obj) == []
        assert security.get_principals(Creator, object=obj) == [user]
        obj.creator = old_creator

        # permissions through group membership
        security.grant_role(group, "manager", obj)
        assert security.has_role(group, "manager", obj)
        assert security.get_roles(group, obj) == ['manager']

        # group membership: user hasn't role set, but has permissions
        assert security.get_roles(user, obj, no_group_roles=True) == []
        assert security.has_permission(user, "read", obj)
        assert security.has_permission(user, "write", obj)
        assert security.has_permission(user, "manage", obj)

        group.members.remove(user)
        self.session.flush()
        assert not security.has_role(user, "manager", obj)
        assert security.get_roles(user, obj) == []
        assert not security.has_permission(user, "read", obj)
        assert not security.has_permission(user, "write", obj)
        assert not security.has_permission(user, "manage", obj)

        security.ungrant_role(group, "manager", obj)
        assert not security.has_role(group, "manager", obj)
        assert security.get_roles(group, obj) == []

        # when called on unmapped instance
        new_obj = DummyModel()
        assert not security.has_permission(user, READ, new_obj)

    def test_grant_roles_unique(self):
        user = User(email="john@example.com", password="x")
        obj = DummyModel()
        self.session.add_all([user, obj])
        self.session.flush()

        assert RoleAssignment.query.count() == 0

        security.grant_role(user, "manager", obj)
        self.session.flush()
        assert RoleAssignment.query.count() == 1

        security.grant_role(user, "manager", obj)
        self.session.flush()
        assert RoleAssignment.query.count() == 1

        security.grant_role(user, "reader", obj)
        self.session.flush()
        assert RoleAssignment.query.count() == 2

    def test_inherit(self):
        folder = FolderishModel()
        self.session.add(folder)
        self.session.flush()

        self.assertEqual(SecurityAudit.query.count(), 0)

        security.set_inherit_security(folder, False)
        self.session.flush()
        self.assertFalse(folder.inherit_security)
        self.assertEqual(SecurityAudit.query.count(), 1)

        security.set_inherit_security(folder, True)
        self.session.flush()
        self.assertTrue(folder.inherit_security)
        self.assertEqual(SecurityAudit.query.count(), 2)

    def test_add_list_delete_permissions(self):
        obj = DummyModel()
        assert security.get_permissions_assignments(obj) == {}
        self.session.add(obj)
        self.session.flush()

        security.add_permission(READ, Authenticated, obj)
        security.add_permission(READ, Owner, obj)
        security.add_permission(WRITE, Owner, obj)
        assert security.get_permissions_assignments(obj) == {
            READ: {Authenticated, Owner},
            WRITE: {Owner},
        }

        security.delete_permission(READ, Authenticated, obj)
        assert security.get_permissions_assignments(obj) == {
            READ: {Owner},
            WRITE: {Owner},
        }
        assert security.get_permissions_assignments(obj, READ) == {
            READ: {Owner},
        }

        # do it twice: it should not crash
        security.add_permission(READ, Owner, obj)
        security.delete_permission(READ, Authenticated, obj)

        # set/get/delete global permission
        security.add_permission(READ, Writer)
        assert security.get_permissions_assignments() == {READ: {Writer,},}

    def test_has_permission_on_objects(self):
        has_permission = security.has_permission
        user = User(email="john@example.com", password="x")
        group = Group(name="Test Group")
        user.groups.add(group)
        obj = DummyModel(creator=user, owner=user)
        self.session.add_all([user, obj])
        self.session.flush()

        # global role provides permissions on any object
        security.grant_role(user, Reader)
        assert has_permission(user, READ, obj=obj)
        assert not has_permission(user, WRITE, obj=obj)

        security.grant_role(user, Writer, obj=obj)
        assert has_permission(user, WRITE, obj=obj)

        # permission assignment
        security.ungrant_role(user, Reader)
        security.ungrant_role(user, Writer, object=obj)
        security.grant_role(user, Authenticated)
        assert not has_permission(user, READ, obj=obj)
        assert not has_permission(user, WRITE, obj=obj)

        pa = PermissionAssignment(
            role=Authenticated, permission=READ, object=obj)
        self.session.add(pa)
        self.session.flush()
        assert has_permission(user, READ, obj=obj)
        assert not has_permission(user, WRITE, obj=obj)

        self.session.delete(pa)
        self.session.flush()
        assert not has_permission(user, READ, obj=obj)

        # Owner / Creator
        for role in (Owner, Creator):
            pa = PermissionAssignment(role=Owner, permission=READ, object=obj)
            self.session.add(pa)
            self.session.flush()
            assert has_permission(user, READ, obj=obj)

            self.session.delete(pa)
            self.session.flush()
            assert not has_permission(user, READ, obj=obj)

        # test when object is *not* in session (newly created objects have id=None
        # for instance)
        obj = DummyModel()
        assert security.has_role(user, Reader, object=obj) is False

    def test_has_permission_custom_roles(self):
        user = User(email="john@example.com", password="x")
        self.session.add(user)
        self.session.flush()

        role = Role('custom_role')
        permission = Permission('custom permission')
        assert not security.has_permission(user, permission, roles=role)
        security.grant_role(user, role)
        assert not security.has_permission(user, permission)
        assert security.has_permission(user, permission, roles=role)

        # Permission always granted if Anonymous role
        assert security.has_permission(user, permission, roles=Anonymous)

        # test convert legacy permission & implicit mapping
        security.grant_role(user, 'reader')
        assert security.has_permission(user, 'read')
        assert not security.has_permission(user, 'write')
        assert not security.has_permission(user, 'manage')

        security.grant_role(user, 'writer')
        assert security.has_permission(user, 'read')
        assert security.has_permission(user, 'write')
        assert not security.has_permission(user, 'manage')

        security.grant_role(user, 'manager')
        assert security.has_permission(user, 'read')
        assert security.has_permission(user, 'write')
        assert security.has_permission(user, 'manage')

    @skip
    def test_query_entity_with_permission(self):
        get_filter = security.query_entity_with_permission
        user = User(email="john@example.com", password="x")
        self.session.add(user)

        obj_reader = DummyModel(name='reader')
        obj_writer = DummyModel(name='writer')
        obj_none = DummyModel(name='none')
        self.session.add_all([obj_reader, obj_writer, obj_none])

        assigments = [
            PermissionAssignment(
                role=Reader, permission=READ, object=obj_reader),
            #
            PermissionAssignment(
                role=Writer, permission=WRITE, object=obj_writer)
        ]
        self.session.add_all(assigments)
        self.session.flush()

        # very unfiltered query returns all objects
        base_query = DummyModel.query
        assert set(base_query.all()) == {obj_reader, obj_writer, obj_none}

        # user has no roles: no objects returned at all
        assert base_query.filter(get_filter(READ, user=user)).all() == []
        assert base_query.filter(get_filter(WRITE, user=user)).all() == []

        # grant object specific roles
        security.grant_role(user, Reader, obj=obj_reader)
        security.grant_role(user, Writer, obj=obj_writer)
        self.session.flush()
        assert base_query.filter(get_filter(READ, user=user)).all() \
            == [obj_reader]
        assert base_query.filter(get_filter(WRITE, user=user)).all() \
            == [obj_writer]

        # Permission granted to anonymous: objects returned
        pa = PermissionAssignment(
            role=Anonymous, permission=WRITE, object=obj_reader)
        self.session.add(pa)

        assert base_query.filter(get_filter(READ, user=user)).all() \
            == [obj_reader]
        assert set(base_query.filter(get_filter(WRITE, user=user)).all()) \
            == {obj_reader, obj_writer}
        self.session.delete(pa)
        assert base_query.filter(get_filter(WRITE, user=user)).all() \
            == [obj_writer]

        # grant global roles
        security.ungrant_role(user, Reader, object=obj_reader)
        security.ungrant_role(user, Writer, object=obj_writer)
        security.grant_role(user, Reader)
        security.grant_role(user, Writer)
        self.session.flush()

        assert base_query.filter(get_filter(
            READ, user=user)).all() == [obj_reader]
        assert base_query.filter(get_filter(
            WRITE, user=user)).all() == [obj_writer]

        # admin role has all permissions
        # 1: local role
        security.ungrant_role(user, Reader)
        security.ungrant_role(user, Writer)
        security.grant_role(user, Admin, obj=obj_reader)
        security.grant_role(user, Admin, obj=obj_none)
        self.session.flush()

        assert set(base_query.filter(get_filter(READ, user=user)).all()) == \
          {obj_reader, obj_none}
        assert set(base_query.filter(get_filter(WRITE, user=user)).all()) == \
          {obj_reader, obj_none}

        # 2: global role
        security.ungrant_role(user, Admin, object=obj_reader)
        security.ungrant_role(user, Admin, object=obj_none)
        security.grant_role(user, Admin)
        self.session.flush()
        assert set(base_query.filter(get_filter(READ, user=user)).all()) == \
            {obj_reader, obj_writer, obj_none}

        # implicit role: Owner, Creator
        security.ungrant_role(user, Admin)
        assert base_query.filter(get_filter(READ, user=user)).all() == []
        assert base_query.filter(get_filter(WRITE, user=user)).all() == []

        obj_reader.creator = user
        obj_writer.owner = user
        assigments = [
            PermissionAssignment(
                role=Creator, permission=READ, object=obj_reader),
            #
            PermissionAssignment(
                role=Owner, permission=WRITE, object=obj_writer)
        ]
        self.session.add_all(assigments)
        self.session.flush()

        assert base_query.filter(get_filter(READ, user=user)).all() \
            == [obj_reader]
        assert base_query.filter(get_filter(WRITE, user=user)).all() \
            == [obj_writer]


class PermissionNoSAWarnTestCase(IntegrationTestCase):

    SQLALCHEMY_WARNINGS_AS_ERROR = False

    def test_add_delete_permissions_expunged_obj(self):
        # weird case. In CreateObject based views, usually Entity is instanciated
        # and might be added to session if it has a relationship with an existing
        # object. `init_object` must do `session.expunge(obj)`. But entities will
        # have initialized default permissions during `after_attach`.
        #
        # At save time, the object is added again to session. The bug is that
        # without precaution we may create permissions assignment twice, because
        # assignments created in the first place are not yet again in session.(new,
        # dirty, deleted) and cannot be found with a filtered query on
        # PermissionAssignment because they have not been flushed yet.
        #
        security.add_permission(READ, Owner, None)
        obj = DummyModel()
        # override default permission at instance level
        obj.__default_permissions__ = frozenset({(READ, frozenset({Owner})),})
        # core.entities._setup_default_permissions creates
        self.session.add(obj)
        # permissions
        security.add_permission(READ, Owner, obj)  # no-op
        # obj and its permissions are removed from session
        self.session.expunge(obj)

        self.session.add(obj)  # obj in session again. When
        # _setup_default_permissions is called durint
        # `after_flush`, previously created permission are not
        # yet back in session. The cascading rule will add
        # them just after (as of sqlalchemy 0.8, at least)

        # Finally the test! IntegrityError will be raised if we have done something
        # wrong (`Key (permission, role, object_id)=(..., ..., ...) already exists`)
        self.session.flush()



from __future__ import absolute_import, division, print_function, \
    unicode_literals

# TODO: most of this is currently only stubs and needs to be implemented.
#
# Validators
#
# NOTE: the `rule` property is supposed to be useful for generating client-side
# validation code.
from wtforms.compat import string_types
from wtforms.validators import URL, UUID, AnyOf, DataRequired, Email, \
    EqualTo, IPAddress, MacAddress, NoneOf, NumberRange, Optional, Regexp, \
    StopValidation, ValidationError

from abilian.i18n import _, _n
from abilian.services import get_service


class Rule(object):

    @property
    def rule(self):
        return None


class Email(Email):

    def __call__(self, form, field):
        if self.message is None:
            self.message = field.gettext(u'Invalid email address.')

        if field.data:
            super(Email, self).__call__(form, field)

    @property
    def rule(self):
        return {"email": True}


class CorrectInputRequired(DataRequired):

    def __call__(self, form, field):
        if field.data is None or (isinstance(field.data, string_types) and
                                  not field.data.strip()) or (
                                      isinstance(field.data, (list, dict)) and
                                      not field.data):
            if self.message is None:
                message = field.gettext('This field is required.')
            else:
                message = self.message

            field.errors[:] = []
            raise StopValidation(message)


class Required(CorrectInputRequired):
    field_flags = ('required',)

    @property
    def rule(self):
        return {"required": True}


class EqualTo(EqualTo, Rule):
    pass


class Length(Rule):
    """Validates the length of a string.

    :param min: The minimum required length of the string. If not provided, minimum
    length will not be checked.

    :param max:  The maximum length of the string. If not provided, maximum length
    will not be checked.

    :param message: Error message to raise in case of a validation error. Can be
    interpolated using `%(min)d` and `%(max)d` if desired. Useful defaults are
    provided depending on the existence of min and max.
    """

    def __init__(self, min=-1, max=-1, message=None):
        assert min != -1 or max != -1, 'At least one of `min` or `max` must be specified.'
        assert max == -1 or min <= max, '`min` cannot be more than `max`.'
        self.min = min
        self.max = max
        self.message = message

    def __call__(self, form, field):
        l = field.data and len(field.data) or 0

        if l < self.min or self.max != -1 and l > self.max:
            message = self.message
            if message is None:
                if self.max == -1:
                    message = _n(
                        u'Field must be at least %(min)d character long.',
                        u'Field must be at least %(min)d characters long.',
                        self.min,
                        min=self.min)
                elif self.min == -1:
                    message = _n(
                        u'Field cannot be longer than %(max)d character.',
                        u'Field cannot be longer than %(max)d characters.',
                        self.max,
                        max=self.max)
                else:
                    message = _(
                        u'Field must be between %(min)d and %(max)d characters long.',
                        min=self.min, max=self.max)
            raise ValidationError(message % dict(
                min=self.min, max=self.max, length=l))


class NumberRange(NumberRange, Rule):
    pass


class Optional(Optional, Rule):
    pass


class Regexp(Regexp, Rule):
    pass


class IPAddress(IPAddress, Rule):
    pass


class MacAddress(MacAddress, Rule):
    pass


class URL(URL):

    @property
    def rule(self):
        return {"url": True}


class UUID(UUID, Rule):
    pass


class AnyOf(AnyOf, Rule):
    pass


class NoneOf(NoneOf, Rule):
    pass


class FlagHidden(Rule):
    """Flag the field as hidden.
    """
    field_flags = ('hidden',)

    def __call__(self, form, field):
        pass


class AntiVirus(Rule):
    """Check content for viruses.
    """
    field_flags = ('antivirus',)

    def __call__(self, form, field):
        svc = get_service('antivirus')
        if not svc:
            return

        res = svc.scan(field.data)
        if res is False:
            raise ValidationError(_(u'Virus detected!'))


class RenderEmpty(object):
    """Force display.
    """
    field_flags = ('render_empty',)

    def __call__(self, form, field):
        pass


# These are the canonical names that should be used.
equalto = EqualTo
length = Length
numberrange = NumberRange
optional = Optional
required = Required
regexp = Regexp
email = Email
ipaddress = IPAddress
macaddress = MacAddress
url = URL
uuid = UUID
anyof = AnyOf
noneof = NoneOf
flaghidden = FlagHidden
renderempty = RenderEmpty



"""Another try at an activity service.

See: ActivityStreams, OpenSocial

Note: ICOM also a notion of activity, but it's very different.
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals


class Activity(object):
    """
  Persistent object that represents an activity entry.
  """


class ActivityService(object):

    def post(self, activity):
        """Post an activity to the service.
    """

    # Now we need some methods to query the stream



from __future__ import absolute_import, division, print_function, \
    unicode_literals

from sqlalchemy import Column, Text, UnicodeText

from abilian.core.entities import SEARCHABLE, Entity


class DummyContact(Entity):
    salutation = Column(UnicodeText, default="")
    first_name = Column(UnicodeText, default="", info=SEARCHABLE)
    last_name = Column(UnicodeText, default="", info=SEARCHABLE)
    email = Column(Text, default="")



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from flask import g
from six import text_type
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship
from sqlalchemy.schema import Column, ForeignKey
from whoosh.fields import STORED

from .base import AUDITABLE, EDITABLE, SEARCHABLE, SYSTEM
from .subjects import User


class OwnedMixin(object):
    __indexation_args__ = {
        'index_to': (
            ('creator', ('creator',)),
            ('creator_name', (('creator_name', STORED),)),
            ('owner', ('owner',)),
            ('owner_name', (('owner_name', STORED),)),)
    }

    def __init__(self, *args, **kwargs):
        try:
            user = g.user
            if not self.creator and not g.user.is_anonymous:
                self.creator = user
            if not self.owner and not g.user.is_anonymous:
                self.owner = user
        except (RuntimeError, AttributeError):
            pass

    @declared_attr
    def creator_id(cls):
        return Column(ForeignKey(User.id), info=SYSTEM)

    @declared_attr
    def creator(cls):
        pj = "User.id == %s.creator_id" % cls.__name__
        return relationship(
            User,
            primaryjoin=pj,
            lazy='joined',
            uselist=False,
            info=SYSTEM | SEARCHABLE)

    @property
    def creator_name(self):
        return text_type(self.creator) if self.creator else u''

    @declared_attr
    def owner_id(cls):
        return Column(ForeignKey(User.id), info=EDITABLE | AUDITABLE)

    @declared_attr
    def owner(cls):
        pj = "User.id == %s.owner_id" % cls.__name__
        return relationship(
            User,
            primaryjoin=pj,
            lazy='joined',
            uselist=False,
            info=EDITABLE | AUDITABLE | SEARCHABLE)

    @property
    def owner_name(self):
        return text_type(self.owner) if self.owner else u''



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, division

from .base import View, JSONView
from .registry import Registry, default_view
from .object import (BaseObjectView, ObjectView, ObjectEdit, ObjectCreate,
                     ObjectDelete, JSONBaseSearch, JSONModelSearch,
                     JSONWhooshSearch)

__all__ = [
    'View',
    'JSONView',
    'Registry',
    'default_view',
    'BaseObjectView',
    'ObjectView',
    'ObjectEdit',
    'ObjectCreate',
    'ObjectDelete',
    'JSONBaseSearch',
    'JSONModelSearch',
    'JSONWhooshSearch',
]



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import pytest
from babel import Locale
from flask import _app_ctx_stack, _request_ctx_stack
from flask_babel import force_locale, get_locale
from jinja2 import DictLoader

from abilian import i18n
from abilian.testing import BaseTestCase

skip = pytest.mark.skip


class I18NTestCase(BaseTestCase):

    @skip
    def test_ensure_request_context(self):
        en = Locale(b'en')
        fr = Locale(b'fr')

        # by default, a request context is present is test case setup.  first let's
        # show no new request context is set up and that set_locale works as
        # advertised
        app_ctx = _app_ctx_stack.top
        current_ctx = _request_ctx_stack.top
        assert current_ctx is not None
        assert get_locale() == en

        with i18n.ensure_request_context():
            assert current_ctx == _request_ctx_stack.top

            with force_locale(fr):
                assert get_locale() == fr

        self._ctx.pop()
        self._ctx = None
        assert _app_ctx_stack.top is None
        assert _request_ctx_stack.top is None
        assert get_locale() is None

        app_ctx.push()

        # # ensure set_locale() doesn't set up a request context
        # with force_locale(fr):
        #     assert _request_ctx_stack.top is None
        #     assert get_locale() is None

        with i18n.ensure_request_context():
            assert _request_ctx_stack.top is not None
            assert current_ctx != _request_ctx_stack.top
            assert get_locale() == en

            with force_locale(fr):
                assert get_locale() == fr

        assert _request_ctx_stack.top is None
        assert get_locale() is None

        # test chaining
        with i18n.ensure_request_context(), force_locale(fr):
            assert get_locale() == fr

    def test_get_template_i18n(self):
        template_path = '/myfile.txt'
        en = Locale('en')
        result = i18n.get_template_i18n(template_path, locale=en)
        self.assertIn('/myfile.en.txt', result)
        self.assertIn('/myfile.txt', result)

        en = Locale('en_US')
        result = i18n.get_template_i18n(template_path, locale=en)
        self.assertIn('/myfile.en_US.txt', result)
        self.assertIn('/myfile.txt', result)

        with force_locale('fr'):
            result = i18n.get_template_i18n(template_path, get_locale())
            self.assertIn('/myfile.fr.txt', result)
            self.assertIn('/myfile.txt', result)

    def test_render_template_i18n(self):
        loader = DictLoader({
            'tmpl.txt': 'default ({{ locale }})',
            'tmpl.en.txt': 'en locale ({{ locale }})',
            'tmpl.fr.txt': 'fr locale ({{ locale }})',
        })
        app_loader = self.app.jinja_loader
        self.app.jinja_loader = loader
        render = i18n.render_template_i18n
        try:
            assert render('tmpl.txt', locale='fr') == 'fr locale (fr)'
            assert render('tmpl.txt', locale='en') == 'en locale (en)'
            assert render('tmpl.txt', locale='de') == 'default (de)'
        finally:
            self.app.jinja_loader = app_loader

    def test_default_country(self):
        assert 'DEFAULT_COUNTRY' in self.app.config
        assert self.app.config['DEFAULT_COUNTRY'] is None
        assert i18n.default_country() is None
        assert i18n.country_choices()[0][0] == 'AF'

        self.app.config['DEFAULT_COUNTRY'] = 'FR'
        assert i18n.default_country() == 'FR'
        assert i18n.country_choices()[0][0] == 'FR'
        assert i18n.country_choices(default_country_first=False)[0][0] == 'AF'



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, division

from .models import Vocabulary  # noqa
from .service import vocabularies, get_vocabulary

__all__ = ['vocabularies']



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from flask import current_app
from flask_login import current_user

from abilian.core.models.subjects import User
from abilian.testing import BaseTestCase

from .models import UserPreference
from .panel import PreferencePanel
from .service import PreferenceService


class VisiblePanel(PreferencePanel):
    id = label = 'visible'

    def is_accessible(self):
        return True

    def get(self):
        return 'Visible'


class AdminPanel(PreferencePanel):
    id = label = 'admin'

    def is_accessible(self):
        security_service = current_app.services['security']
        return security_service.has_role(current_user, "admin")

    def get(self):
        return 'Admin'


class App(BaseTestCase.application_class):

    def init_extensions(self):
        super(App, self).init_extensions()
        prefs = self.services['preferences']
        prefs.app_state.panels = []
        prefs.register_panel(VisiblePanel(), self)
        prefs.register_panel(AdminPanel(), self)


class PreferencesTestCase(BaseTestCase):

    application_class = App

    def test_preferences(self):
        user = User(email="test@example.com")
        assert UserPreference.query.all() == []

        preference_service = PreferenceService()

        preferences = preference_service.get_preferences(user)
        self.assertEqual(preferences, {})

        preference_service.set_preferences(user, digest='daily')
        self.session.flush()

        preferences = preference_service.get_preferences(user)
        self.assertEqual(preferences, {'digest': 'daily'})

        preference_service.clear_preferences(user)
        self.session.flush()

        preferences = preference_service.get_preferences(user)
        self.assertEqual(preferences, {})
        assert UserPreference.query.all() == []

    def test_preferences_with_various_types(self):
        user = User(email="test@example.com")
        preference_service = PreferenceService()

        preference_service.set_preferences(user, some_int=1)
        self.session.flush()
        preferences = preference_service.get_preferences(user)
        self.assertEqual(preferences, {'some_int': 1})

        preference_service.set_preferences(user, some_bool=True)
        self.session.flush()
        preferences = preference_service.get_preferences(user)
        self.assertEqual(preferences, {'some_int': 1, 'some_bool': True})

    def test_visible_panels(self):
        user = User(email="test@example.com")
        app = self.app
        security = app.services['security']
        security.start()

        for cp in app.template_context_processors['preferences']:
            ctx = cp()
            if 'menu' in ctx:
                break

        with self.login(user):
            assert [p['endpoint']
                    for p in ctx['menu']] == ['preferences.visible']
            security.grant_role(user, 'admin')
            ctx = cp()
            assert [p['endpoint'] for p in ctx['menu']
                   ] == ['preferences.visible', 'preferences.admin']



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import logging
from datetime import datetime, timedelta

from flask import current_app, g, redirect, request, url_for
from flask_babel import lazy_gettext as _l
from flask_login import current_user, login_user, user_logged_in, \
    user_logged_out
from werkzeug.exceptions import Forbidden

from abilian.core.extensions import db, login_manager
from abilian.core.models.subjects import User
from abilian.core.signals import user_loaded
from abilian.services import Service, ServiceState
from abilian.web.action import DynamicIcon, actions
from abilian.web.nav import NavGroup, NavItem

from .models import LoginSession
from .views import login as login_views

__all__ = ['AuthService', 'user_menu']

logger = logging.getLogger(__name__)


def is_anonymous(context):
    return current_user.is_anonymous


def is_authenticated(context):
    return not is_anonymous(context)


def _user_photo_endpoint():
    from abilian.web.views import images  # late import: avoid circular import
    return images.user_url_args(current_user, 16)[0]


def _user_photo_icon_args(icon, url_args):
    from abilian.web.views import images  # late import avoid circular import
    return images.user_url_args(current_user, max(icon.width, icon.height))[1]


user_menu = NavGroup(
    'user',
    'authenticated',
    title=lambda c: current_user.name,
    icon=DynamicIcon(
        endpoint=_user_photo_endpoint,
        css='avatar',
        size=20,
        url_args=_user_photo_icon_args),
    condition=is_authenticated,
    items=(NavItem(
        'user',
        'logout',
        title=_l(u'Logout'),
        icon='log-out',
        url=lambda context: url_for('login.logout'),
        divider=True),))

_ACTIONS = (
    NavItem(
        'user',
        'login',
        title=_l(u'Login'),
        icon='log-in',
        url=lambda context: url_for('login.login_form'),
        condition=is_anonymous),
    user_menu,)


class AuthServiceState(ServiceState):
    """
    State class for :class:`AuthService`
    """

    def __init__(self, *args, **kwargs):
        super(AuthServiceState, self).__init__(*args, **kwargs)
        self.bp_access_controllers = {None: []}
        self.endpoint_access_controllers = {}

    def add_bp_access_controller(self, blueprint, func):
        self.bp_access_controllers.setdefault(blueprint, []).append(func)

    def add_endpoint_access_controller(self, endpoint, func):
        self.endpoint_access_controllers.setdefault(endpoint, []).append(func)


class AuthService(Service):
    name = 'auth'
    AppStateClass = AuthServiceState

    def init_app(self, app):
        login_manager.init_app(app)
        login_manager.login_view = 'login.login_form'
        Service.init_app(self, app)
        self.login_url_prefix = app.config.get('LOGIN_URL', '/user')
        app.before_request(self.do_access_control)
        app.before_request(self.update_user_session_data)
        user_logged_in.connect(self.user_logged_in, sender=app)
        user_logged_out.connect(self.user_logged_out, sender=app)
        app.register_blueprint(login_views, url_prefix=self.login_url_prefix)
        with app.app_context():
            actions.register(*_ACTIONS)

    @login_manager.user_loader
    def load_user(user_id):
        try:
            user = User.query.get(user_id)

            if not user or not user.can_login:
                # if a user is edited and should not have access any more, this will
                # ensure he cannot continue if he had an active session
                return None
        except:
            logger.warning("Error during login.", exc_info=True)
            session = current_app.db.session()
            if not session.is_active:
                # session is not usable, rollback should restore a usable session
                session.rollback()

            return None

        app = current_app._get_current_object()
        app.services[AuthService.name].user_logged_in(app, user)
        user_loaded.send(app, user=user)
        return user

    def user_logged_in(self, app, user):
        # `g.user` is used as `current_user`, but `current_user` is actually looking
        # for `request.user` whereas `g` is on app local stack.
        #
        # `g.logged_user` is the actual user. In the future for example we may allow
        # a manager to see site as another user (impersonate), or propose a "see as
        # anonymous" function
        g.user = g.logged_user = user
        is_anonymous = user is None or user.is_anonymous
        security = app.services.get('security')
        g.is_manager = (user and not is_anonymous and (
            (security.has_role(user, 'admin') or
             security.has_role(user, 'manager'))))

    def user_logged_out(self, app, user):
        if hasattr(g, 'user'):
            del g.user
            del g.logged_user
            del g.is_manager

    def redirect_to_login(self, next_url=True):
        kw = {}
        if next_url is not False:
            kw['next'] = request.url if next_url is True else next_url

        return redirect(url_for(login_manager.login_view, **kw))

    def do_access_control(self):
        """
        `before_request` handler to check if user should be redirected to login
        page.
        """
        if current_app.testing and current_app.config.get("NO_LOGIN"):
            # Special case for tests
            user = User.query.get(0)
            login_user(user, False, True)
            return

        state = self.app_state
        user = current_user._get_current_object()
        roles = frozenset(current_app.services['security'].get_roles(user))
        endpoint = request.endpoint
        bp = request.blueprint

        controllers = []
        controllers.extend(state.bp_access_controllers.get(None, []))

        if bp and bp in state.bp_access_controllers:
            controllers.extend(state.bp_access_controllers[bp])

        if endpoint and endpoint in state.endpoint_access_controllers:
            controllers.extend(state.endpoint_access_controllers[endpoint])

        for ctrl in reversed(controllers):
            verdict = ctrl(user=user, roles=roles)
            if verdict is None:
                continue
            elif verdict is True:
                return
            else:
                if user.is_anonymous:
                    return self.redirect_to_login()
                raise Forbidden()

        # default policy
        if current_app.config.get('PRIVATE_SITE') and user.is_anonymous:
            return self.redirect_to_login()

    def update_user_session_data(self):
        user = current_user
        if current_user.is_anonymous:
            return

        # Update last_active every 60 seconds only so as to not stress the database
        # too much.
        now = datetime.utcnow()
        if (user.last_active is None or
            (now - user.last_active) > timedelta(minutes=1)):
            user.last_active = now
            db.session.add(user)
            db.session.commit()

        refresh_login_session(user)


def refresh_login_session(user):
    now = datetime.utcnow()
    session = LoginSession.query.get_active_for(user)
    if not session:
        return

    from_now = now - session.last_active_at

    if from_now > timedelta(hours=1):
        session.ended_at = session.last_active_at
        session = LoginSession.new()
        db.session.add(session)
        db.session.commit()
    elif from_now > timedelta(minutes=1):
        session.last_active_at = now
        db.session.commit()



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from six import python_2_unicode_compatible, text_type, with_metaclass
from sqlalchemy.types import String, TypeDecorator


class ValueSingletonMeta(type):

    def __new__(cls, name, bases, dct):
        dct['__instances__'] = {}
        dct.setdefault('__slots__', ())
        new_type = type.__new__(cls, name, bases, dct)
        return new_type

    def __call__(cls, value, *args, **kwargs):
        if isinstance(value, cls):
            return value

        if value not in cls.__instances__:
            value_instance = type.__call__(cls, value, *args, **kwargs)
            cls.__instances__[getattr(value_instance,
                                      cls.attr)] = value_instance
        return cls.__instances__[value.lower()]


@python_2_unicode_compatible
class UniqueName(with_metaclass(ValueSingletonMeta, object)):
    """Base class to create singletons from strings.

    A subclass of :class:`UniqueName` defines a namespace.
    """
    # __metaclass__ = ValueSingletonMeta
    __slots__ = ('_hash', '__name')
    attr = 'name'

    def __init__(self, name):
        self.__name = text_type(name).strip().lower()
        self._hash = hash(self.__name)

    @property
    def name(self):
        return self.__name

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, repr(self.name))

    def __str__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._hash == other._hash
        return self.__name == text_type(other)

    def __hash__(self):
        return self._hash


class UniqueNameType(TypeDecorator):
    """Sqlalchemy type that stores a subclass of :class:`UniqueName`

    Usage::
    class MySingleton(UniqueName):
        pass

    class MySingletonType(UniqueNameType):
        Type = MySingleton
    """
    impl = String
    Type = None
    default_max_length = 100

    def __init__(self, *args, **kwargs):
        assert self.Type is not None
        kwargs.setdefault('length', self.default_max_length)
        TypeDecorator.__init__(self, *args, **kwargs)

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = str(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = self.Type(value)
        return value



"""
All signals used by Abilian Core.

Signals are the main tools used for decoupling applications components by
sending notifications. In short, signals allow certain senders to notify
subscribers that something happened.

Cf. http://flask.pocoo.org/docs/signals/ for detailed documentation.

The main signal is currently :obj:`activity`.
"""

from __future__ import absolute_import, division, print_function, \
    unicode_literals

from blinker.base import Namespace

signals = Namespace()

#: Triggered at application initialization when all extensions and plugins have
#: been loaded
components_registered = signals.signal("app:components:registered")

#: Trigger when JS api must be registered. At this time :func:`flask.url_for` is
#: usable
register_js_api = signals.signal('app:register-js-api')

#: This signal is used by the activity streams service and its clients.
activity = signals.signal("activity")

#: This signal is sent when user object has been loaded. g.user and current_user
#: are available.
user_loaded = signals.signal('user_loaded')



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from wtforms.fields import StringField

from abilian.core.models.attachment import Attachment
from abilian.i18n import _l
from abilian.web.forms import Form
from abilian.web.forms.fields import FileField
from abilian.web.forms.filters import strip
from abilian.web.forms.validators import required


class AttachmentForm(Form):

    blob = FileField(
        _l(u'file'), validators=[required()], filters=[strip], multiple=False)

    description = StringField(_l(u'description (optional)'), filters=[strip])

    class Meta:
        model = Attachment
        include_primary_keys = True
        assign_required = False  # for 'id': allow None, for new records



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import sqlalchemy as sa
from flask import Blueprint, Flask
from flask_testing import TestCase as FlaskTestCase

from abilian.core.entities import Entity
# needed if running only this test, else SA won't have registered this mapping
# required by Entity.owner_id, etc
from abilian.core.models.subjects import User  # noqa
from abilian.web.views import Registry, default_view


class RegEntity(Entity):
    name = sa.Column(sa.Unicode, default='')


class NonEntity(object):
    pass


class TestRegistry(FlaskTestCase):

    def create_app(self):
        app = Flask(__name__)
        app.config.update({
            'TESTING': True,
            'CSRF_ENABLED': False,
            'WTF_CSRF_ENABLED': False,
        })

        app.default_view = Registry()
        return app

    def test_register_class_or_instance(self):
        obj = RegEntity()
        registry = self.app.default_view

        registry.register(RegEntity, lambda ignored: '')
        assert RegEntity.entity_type in registry._map
        registry._map = {}
        registry.register(obj, lambda ignored: '')
        assert RegEntity.entity_type in registry._map

    def test_custom_url_func(self):
        name = u'obj'
        obj = RegEntity(id=1, name=name)
        registry = self.app.default_view

        def custom_url(obj, obj_type, obj_id):
            return obj.name

        registry.register(obj, custom_url)
        self.assertEqual(registry.url_for(obj), name)

        def url_from_type_and_id(obj, obj_type, obj_id):
            return u'{}:{}'.format(obj_type, obj_id)

        registry.register(obj, url_from_type_and_id)
        self.assertEqual(registry.url_for(obj), u'test_registry.RegEntity:1')

    def test_default_url_func(self):
        obj = RegEntity(id=1)

        @self.app.route(
            '/regentities_path/<int:object_id>/view', endpoint='regentity.view')
        def dummy_default_view(object_id):
            pass

        self.assertEqual(
            self.app.default_view.url_for(obj), '/regentities_path/1/view')

        self.assertEqual(
            self.app.default_view.url_for(
                obj, _external=True),
            'http://localhost/regentities_path/1/view')

    def test_default_view_decorator(self):
        bp = Blueprint('registry', __name__, url_prefix='/blueprint')

        @default_view(bp, RegEntity)
        @bp.route('/<int:object_id>')
        def view(object_id):
            pass

        obj = RegEntity(id=1)
        # blueprint not registered: no rule set
        self.assertRaises(KeyError, self.app.default_view.url_for, obj)

        # blueprint registered: default view is set
        self.app.register_blueprint(bp)
        self.assertEqual(self.app.default_view.url_for(obj), '/blueprint/1')
        self.assertEqual(
            self.app.default_view.url_for(
                obj, _external=True),
            'http://localhost/blueprint/1')
